# Copyright 2019 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import pytest

from pex.vendor import VendorSpec


def test_pinned():
  vendor_spec = VendorSpec.pinned('foo', '1.2.3')
  assert 'foo' == vendor_spec.key
  assert 'foo==1.2.3' == vendor_spec.requirement


def test_vcs_valid():
  vendor_spec = VendorSpec.vcs('git+https://github.com/foo.git@da39a3ee#egg=bar')
  assert 'bar' == vendor_spec.key
  assert 'git+https://github.com/foo.git@da39a3ee#egg=bar' == vendor_spec.requirement


def test_vcs_invalid_no_egg():
  with pytest.raises(ValueError):
    VendorSpec.vcs('git+https://github.com/foo.git@da39a3ee')


def test_vcs_invalid_multiple_egg():
  with pytest.raises(ValueError):
    VendorSpec.vcs('git+https://github.com/foo.git@da39a3ee#egg=bar&egg=foo')



# Copyright 2019 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import json
import os
import subprocess
from subprocess import CalledProcessError
from textwrap import dedent

import pytest

from pex.common import safe_mkdir, safe_mkdtemp, safe_open, safe_rmtree
from pex.pex_info import PexInfo
from pex.pip import get_pip
from pex.testing import run_pex_command


@pytest.fixture(scope='module')
def pex_project_dir():
  return subprocess.check_output(['git', 'rev-parse', '--show-toplevel']).decode('utf-8').strip()


def test_issues_789_demo(pex_project_dir):
  tmpdir = safe_mkdtemp()

  # 1. Imagine we've pre-resolved the requirements needed in our wheel house.
  requirements = [
    'ansicolors',
    'isort',
    'setuptools'  # N.B.: isort doesn't declare its setuptools dependency.
  ]

  wheelhouse = os.path.join(tmpdir, 'wheelhouse')
  get_pip().spawn_download_distributions(
    download_dir=wheelhouse,
    requirements=requirements
  ).wait()

  # 2. Also imagine this configuration is passed to a tool (PEX or a wrapper as in this test
  # example) via the CLI or other configuration data sources. For example, Pants has a `PythonSetup`
  # that combines with BUILD target data to get you this sort of configuration info outside pex.
  resolver_settings = dict(
    indexes=[],  # Turn off pypi.
    find_links=[wheelhouse],  # Use our wheel house.
    build=False,  # Use only pre-built wheels.
  )

  # 3. That same configuration was used to build a standard pex:
  resolver_args = []
  if len(resolver_settings['find_links']) == 0:
    resolver_args.append('--no-index')
  else:
    for index in resolver_settings['indexes']:
      resolver_args.extend(['--index', index])

  for repo in resolver_settings['find_links']:
    resolver_args.extend(['--find-links', repo])

  resolver_args.append('--build' if resolver_settings['build'] else '--no-build')

  project_code_dir = os.path.join(tmpdir, 'project_code_dir')
  with safe_open(os.path.join(project_code_dir, 'colorized_isort.py'), 'w') as fp:
    fp.write(dedent("""\
      import colors
      import os
      import subprocess
      import sys


      def run():
        env = os.environ.copy()
        env.update(PEX_MODULE='isort')
        isort_process = subprocess.Popen(
          sys.argv,
          env=env,
          stdout = subprocess.PIPE,
          stderr = subprocess.PIPE
        )
        stdout, stderr = isort_process.communicate()
        print(colors.green(stdout.decode('utf-8')))
        print(colors.red(stderr.decode('utf-8')))
        sys.exit(isort_process.returncode)
    """))

  colorized_isort_pex = os.path.join(tmpdir, 'colorized_isort.pex')
  args = [
    '--sources-directory', project_code_dir,
    '--entry-point', 'colorized_isort:run',
    '--output-file', colorized_isort_pex
  ]
  result = run_pex_command(args + resolver_args + requirements)
  result.assert_success()

  # 4. Now the tool builds a "dehydrated" PEX using the standard pex + resolve settings as the
  # template.
  ptex_cache = os.path.join(tmpdir, '.ptex')

  colorized_isort_pex_info = PexInfo.from_pex(colorized_isort_pex)
  colorized_isort_pex_info.pex_root = ptex_cache

  # Force the standard pex to extract its code. An external tool like Pants would already know the
  # orignal source code file paths, but we need to discover here.
  colorized_isort_pex_code_dir = os.path.join(
    colorized_isort_pex_info.zip_unsafe_cache,
    colorized_isort_pex_info.code_hash
  )
  env = os.environ.copy()
  env.update(PEX_ROOT=ptex_cache, PEX_INTERPRETER='1', PEX_FORCE_LOCAL='1')
  subprocess.check_call([colorized_isort_pex, '-c', ''], env=env)

  colorized_isort_ptex_code_dir = os.path.join(tmpdir, 'colorized_isort_ptex_code_dir')
  safe_mkdir(colorized_isort_ptex_code_dir)

  code = []
  for root, dirs, files in os.walk(colorized_isort_pex_code_dir):
    rel_root = os.path.relpath(root, colorized_isort_pex_code_dir)
    for f in files:
      # Don't ship compiled python from the code extract above, the target interpreter will not
      # match ours in general.
      if f.endswith('.pyc'):
        continue
      rel_path = os.path.normpath(os.path.join(rel_root, f))
      # The root __main__.py is special for any zipapp including pex, let it write its own
      # __main__.py bootstrap. Similarly. PEX-INFO is special to pex and we want the PEX-INFO for
      # The ptex pex, not the pex being ptexed.
      if rel_path in ('__main__.py', PexInfo.PATH):
        continue
      os.symlink(os.path.join(root, f), os.path.join(colorized_isort_ptex_code_dir, rel_path))
      code.append(rel_path)

  ptex_code_dir = os.path.join(tmpdir, 'ptex_code_dir')

  ptex_info = dict(code=code, resolver_settings=resolver_settings)
  with safe_open(os.path.join(ptex_code_dir, 'PTEX-INFO'), 'w') as fp:
    json.dump(ptex_info, fp)

  with safe_open(os.path.join(ptex_code_dir, 'IPEX-INFO'), 'w') as fp:
    fp.write(colorized_isort_pex_info.dump())

  with safe_open(os.path.join(ptex_code_dir, 'ptex.py'), 'w') as fp:
    fp.write(dedent("""\
      import json
      import os
      import sys

      from pex import resolver
      from pex.common import open_zip
      from pex.pex_builder import PEXBuilder
      from pex.pex_info import PexInfo
      from pex.util import CacheHelper
      from pex.variables import ENV

      self = sys.argv[0]
      ipex_file = '{}.ipex'.format(os.path.splitext(self)[0])

      if not os.path.isfile(ipex_file):
        print('Hydrating {} to {}'.format(self, ipex_file))

        ptex_pex_info = PexInfo.from_pex(self)
        code_root = os.path.join(ptex_pex_info.zip_unsafe_cache, ptex_pex_info.code_hash)
        with open_zip(self) as zf:
          # Populate the pex with the pinned requirements and distribution names & hashes.
          ipex_info = PexInfo.from_json(zf.read('IPEX-INFO'))
          ipex_builder = PEXBuilder(pex_info=ipex_info)

          # Populate the pex with the needed code.
          ptex_info = json.loads(zf.read('PTEX-INFO').decode('utf-8'))
          for path in ptex_info['code']:
            ipex_builder.add_source(os.path.join(code_root, path), path)

        # Perform a fully pinned intransitive resolve to hydrate the install cache (not the
        # pex!).
        resolver_settings = ptex_info['resolver_settings']
        resolved_distributions = resolver.resolve(
          requirements=[str(req) for req in ipex_info.requirements],
          cache=ipex_info.pex_root,
          transitive=False,
          **resolver_settings
        )

        ipex_builder.build(ipex_file)

      os.execv(ipex_file, [ipex_file] + sys.argv[1:])
    """))

  colorized_isort_ptex = os.path.join(tmpdir, 'colorized_isort.ptex')

  result = run_pex_command([
    '--not-zip-safe',
    '--always-write-cache',
    '--pex-root', ptex_cache,
    pex_project_dir,
    '--sources-directory', ptex_code_dir,
    '--sources-directory', colorized_isort_ptex_code_dir,
    '--entry-point', 'ptex',
    '--output-file', colorized_isort_ptex
  ])
  result.assert_success()

  subprocess.check_call([colorized_isort_ptex, '--version'])
  with pytest.raises(CalledProcessError):
    subprocess.check_call([colorized_isort_ptex, '--not-a-flag'])

  safe_rmtree(ptex_cache)

  # The dehydrated pex now fails since it lost its hydration from the cache.
  with pytest.raises(CalledProcessError):
    subprocess.check_call([colorized_isort_ptex, '--version'])



# Copyright 2017 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

# A library of functions for filtering Python interpreters based on compatibility constraints

from __future__ import absolute_import

from pex.common import die
from pex.interpreter import PythonIdentity
from pex.tracer import TRACER


def validate_constraints(constraints):
  # TODO: add check to see if constraints are mutually exclusive (bad) so no time is wasted:
  # https://github.com/pantsbuild/pex/issues/432
  for req in constraints:
    # Check that the compatibility requirements are well-formed.
    try:
      PythonIdentity.parse_requirement(req)
    except ValueError as e:
      die("Compatibility requirements are not formatted properly: %s" % str(e))


class UnsatisfiableInterpreterConstraintsError(Exception):
  """Indicates interpreter constraints could not be satisfied."""

  def __init__(self, constraints, candidates):
    """
    :param constraints: The constraints that could not be satisfied.
    :type constraints: iterable of str
    :param candidates: The python interpreters that were compared against the constraints.
    :type candidates: iterable of :class:`pex.interpreter.PythonInterpreter`
    """
    self.constraints = tuple(constraints)
    self.candidates = tuple(candidates)
    super(UnsatisfiableInterpreterConstraintsError, self).__init__(self.create_message())

  def create_message(self, preamble=None):
    """Create a message describing  failure to find matching interpreters with an optional preamble.

    :param str preamble: An optional preamble to the message that will be displayed above it
                         separated by an empty blank line.
    :return: A descriptive message useable for display to an end user.
    :rtype: str
    """
    binary_column_width = max(len(candidate.binary) for candidate in self.candidates)
    interpreters_format = '{{binary: >{}}} {{requirement}}'.format(binary_column_width)
    return (
      '{preamble}'
      'Examined the following interpreters:\n  {interpreters}\n\n'
      'None were compatible with the requested constraints:\n  {constraints}'
    ).format(
      preamble='{}\n\n'.format(preamble) if preamble else '',
      interpreters='\n  '.join(interpreters_format.format(
        binary=candidate.binary,
        requirement=candidate.identity.requirement
      ) for candidate in self.candidates),
      constraints='\n  '.join(self.constraints)
    )


def matched_interpreters_iter(interpreters_iter, constraints):
  """Given some filters, yield any interpreter that matches at least one of them.

  :param interpreters_iter: A `PythonInterpreter` iterable for filtering.
  :param constraints: A sequence of strings that constrain the interpreter compatibility for this
    pex. Each string uses the Requirement-style format, e.g. 'CPython>=3' or '>=2.7,<3' for
    requirements agnostic to interpreter class. Multiple requirement strings may be combined
    into a list to OR the constraints, such as ['CPython>=2.7,<3', 'CPython>=3.4'].
  :return: returns a generator that yields compatible interpreters
  :raises: :class:`UnsatisfiableInterpreterConstraintsError` if constraints were given and could
           not be satisfied. The exception will only be raised when the returned generator is fully
           consumed.
  """
  candidates = []
  found = False

  for interpreter in interpreters_iter:
    if any(interpreter.identity.matches(filt) for filt in constraints):
      TRACER.log("Constraints on interpreters: %s, Matching Interpreter: %s"
                 % (constraints, interpreter.binary), V=3)
      found = True
      yield interpreter

    if not found:
      candidates.append(interpreter)

  if not found:
    raise UnsatisfiableInterpreterConstraintsError(constraints, candidates)



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import warnings

import pytest

from pex.common import temporary_dir
from pex.pex_warnings import PEXWarning
from pex.testing import environment_as
from pex.util import named_temporary_file
from pex.variables import Variables


def test_process_pydoc():
  def thing():
    # no pydoc
    pass
  assert Variables.process_pydoc(thing.__doc__) == ('Unknown', 'Unknown')

  def other_thing():
    """Type

    Properly
         formatted
      text.
    """

  assert Variables.process_pydoc(other_thing.__doc__) == (
      'Type', 'Properly formatted text.')


def test_iter_help():
  for variable_name, variable_type, variable_text in Variables.iter_help():
    assert variable_name.startswith('PEX_')
    assert '\n' not in variable_type
    assert '\n' not in variable_text


def test_pex_bool_variables():
  Variables(environ={})._get_bool('NOT_HERE', default=False) is False
  Variables(environ={})._get_bool('NOT_HERE', default=True) is True

  for value in ('0', 'faLsE', 'false'):
    for default in (True, False):
      Variables(environ={'HERE': value})._get_bool('HERE', default=default) is False
  for value in ('1', 'TrUe', 'true'):
    for default in (True, False):
      Variables(environ={'HERE': value})._get_bool('HERE', default=default) is True
  with pytest.raises(SystemExit):
    Variables(environ={'HERE': 'garbage'})._get_bool('HERE')

  # end to end
  assert Variables().PEX_ALWAYS_CACHE is False
  assert Variables({'PEX_ALWAYS_CACHE': '1'}).PEX_ALWAYS_CACHE is True


def test_pex_string_variables():
  Variables(environ={})._get_string('NOT_HERE') is None
  Variables(environ={})._get_string('NOT_HERE', default='lolol') == 'lolol'
  Variables(environ={'HERE': 'stuff'})._get_string('HERE') == 'stuff'
  Variables(environ={'HERE': 'stuff'})._get_string('HERE', default='lolol') == 'stuff'


def test_pex_get_int():
  assert Variables()._get_int('HELLO') is None
  assert Variables()._get_int('HELLO', default=42) == 42
  assert Variables(environ={'HELLO': 23})._get_int('HELLO') == 23
  assert Variables(environ={'HELLO': 23})._get_int('HELLO', default=42) == 23

  with pytest.raises(SystemExit):
    assert Variables(environ={'HELLO': 'welp'})._get_int('HELLO')


def assert_pex_vars_hermetic():
  v = Variables()
  assert os.environ == v.copy()

  existing = os.environ.get('TEST')
  expected = (existing or '') + 'different'
  assert expected != existing

  with environment_as(TEST=expected):
    assert expected != v.copy().get('TEST')


def test_pex_vars_hermetic_no_pexrc():
  assert_pex_vars_hermetic()


def test_pex_vars_hermetic():
  with environment_as(PEX_IGNORE_RCFILES='True'):
    assert_pex_vars_hermetic()


def test_pex_get_kv():
  v = Variables(environ={})
  assert v._get_kv('HELLO') is None
  assert v._get_kv('=42') is None
  assert v._get_kv('TOO=MANY=COOKS') is None
  assert v._get_kv('THIS=WORKS') == ['THIS', 'WORKS']


def test_pex_from_rc():
  with named_temporary_file(mode='w') as pexrc:
    pexrc.write('HELLO=42')
    pexrc.flush()
    v = Variables(rc=pexrc.name)
    assert v._get_int('HELLO') == 42


def test_pexrc_precedence():
  with named_temporary_file(mode='w') as pexrc:
    pexrc.write('HELLO=FORTYTWO')
    pexrc.flush()
    v = Variables(rc=pexrc.name, environ={'HELLO': 42})
    assert v._get_int('HELLO') == 42


def test_rc_ignore():
  with named_temporary_file(mode='w') as pexrc:
    pexrc.write('HELLO=FORTYTWO')
    pexrc.flush()
    v = Variables(rc=pexrc.name, environ={'PEX_IGNORE_RCFILES': 'True'})
    assert 'HELLO' not in v._environ


def test_pex_vars_defaults_stripped():
  v = Variables(environ={})
  stripped = v.strip_defaults()

  # bool
  assert v.PEX_ALWAYS_CACHE is not None
  assert stripped.PEX_ALWAYS_CACHE is None

  # string
  assert v.PEX_PATH is not None
  assert stripped.PEX_PATH is None

  # int
  assert v.PEX_VERBOSE is not None
  assert stripped.PEX_VERBOSE is None


def test_pex_root_unwriteable():
  with temporary_dir() as td:
    pex_root = os.path.realpath(os.path.join(td, "pex_root"))
    os.mkdir(pex_root, 0o444)

    env = Variables(environ=dict(PEX_ROOT=pex_root))

    with warnings.catch_warnings(record=True) as log:
      assert pex_root != env.PEX_ROOT

    assert 1 == len(log)
    message = log[0].message
    assert isinstance(message, PEXWarning)
    assert pex_root in str(message)
    assert env.PEX_ROOT in str(message)

    assert env.PEX_ROOT == env.PEX_ROOT, (
      "When an ephemeral PEX_ROOT is materialized it should be stable."
    )



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import warnings

import pytest

from pex.common import temporary_dir
from pex.orderedset import OrderedSet
from pex.pex_info import PexInfo
from pex.pex_warnings import PEXWarning
from pex.variables import Variables
from pex.version import __version__ as pex_version


def make_pex_info(requirements):
  return PexInfo(info={'requirements': requirements})


def test_backwards_incompatible_pex_info():
  # forwards compatibility
  pi = make_pex_info(['hello'])
  assert pi.requirements == OrderedSet(['hello'])

  pi = make_pex_info(['hello==0.1', 'world==0.2'])
  assert pi.requirements == OrderedSet(['hello==0.1', 'world==0.2'])

  # malformed
  with pytest.raises(ValueError):
    make_pex_info('hello')

  with pytest.raises(ValueError):
    make_pex_info([('hello', False)])

  # backwards compatibility
  pi = make_pex_info([
      ['hello==0.1', False, None],
      ['world==0.2', False, None],
  ])
  assert pi.requirements == OrderedSet(['hello==0.1', 'world==0.2'])


def assert_same_info(expected, actual):
  assert expected.dump(sort_keys=True) == actual.dump(sort_keys=True)


def test_from_empty_env():
  environ = Variables(environ={})
  info = {}
  assert_same_info(PexInfo(info=info), PexInfo.from_env(env=environ))


def test_from_env():
  with temporary_dir() as td:
    pex_root = os.path.realpath(os.path.join(td, 'pex_root'))
    environ = dict(PEX_ROOT=pex_root,
                   PEX_MODULE='entry:point',
                   PEX_SCRIPT='script.sh',
                   PEX_FORCE_LOCAL='true',
                   PEX_UNZIP='true',
                   PEX_INHERIT_PATH='prefer',
                   PEX_IGNORE_ERRORS='true',
                   PEX_ALWAYS_CACHE='true')

    info = dict(pex_root=pex_root,
                entry_point='entry:point',
                script='script.sh',
                zip_safe=False,
                unzip=True,
                inherit_path=True,
                ignore_errors=True,
                always_write_cache=True)

  assert_same_info(PexInfo(info=info), PexInfo.from_env(env=Variables(environ=environ)))


def test_build_properties():
  assert pex_version == PexInfo.default().build_properties['pex_version']


def test_merge_split():
  path_1, path_2 = '/pex/path/1:/pex/path/2', '/pex/path/3:/pex/path/4'
  result = PexInfo._merge_split(path_1, path_2)
  assert result == ['/pex/path/1', '/pex/path/2', '/pex/path/3', '/pex/path/4']

  path_1, path_2 = '/pex/path/1:', '/pex/path/3:/pex/path/4'
  result = PexInfo._merge_split(path_1, path_2)
  assert result == ['/pex/path/1', '/pex/path/3', '/pex/path/4']

  path_1, path_2 = '/pex/path/1::/pex/path/2', '/pex/path/3:/pex/path/4'
  result = PexInfo._merge_split(path_1, path_2)
  assert result == ['/pex/path/1', '/pex/path/2', '/pex/path/3', '/pex/path/4']

  path_1, path_2 = '/pex/path/1::/pex/path/2', '/pex/path/3:/pex/path/4'
  result = PexInfo._merge_split(path_1, None)
  assert result == ['/pex/path/1', '/pex/path/2']
  result = PexInfo._merge_split(None, path_2)
  assert result == ['/pex/path/3', '/pex/path/4']


def test_pex_root_set_none():
  pex_info = PexInfo.default()
  pex_info.pex_root = None

  assert PexInfo.default().pex_root == pex_info.pex_root
  assert os.path.expanduser("~/.pex") == pex_info.pex_root


def test_pex_root_set_unwriteable():
  with temporary_dir() as td:
    pex_root = os.path.realpath(os.path.join(td, "pex_root"))
    os.mkdir(pex_root, 0o444)

    pex_info = PexInfo.default()
    pex_info.pex_root = pex_root

    with warnings.catch_warnings(record=True) as log:
      assert pex_root != pex_info.pex_root

    assert 1 == len(log)
    message = log[0].message
    assert isinstance(message, PEXWarning)
    assert pex_root in str(message)
    assert pex_info.pex_root in str(message)



# coding=utf-8
# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import email

from pex.third_party.packaging.specifiers import SpecifierSet
from pex.third_party.pkg_resources import DistInfoDistribution


def requires_python(dist):
  """Examines dist for `Python-Requires` metadata and returns version constraints if any.

  See: https://www.python.org/dev/peps/pep-0345/#requires-python

  :param dist: A distribution to check for `Python-Requires` metadata.
  :type dist: :class:`pkg_resources.Distribution`
  :return: The required python version specifiers.
  :rtype: :class:`packaging.specifiers.SpecifierSet` or None
  """
  if not dist.has_metadata(DistInfoDistribution.PKG_INFO):
    return None

  metadata = dist.get_metadata(DistInfoDistribution.PKG_INFO)
  pkg_info = email.parser.Parser().parsestr(metadata)
  python_requirement = pkg_info.get('Requires-Python')
  if not python_requirement:
    return None
  return SpecifierSet(python_requirement)



# Copyright 2016 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess
from contextlib import contextmanager
from textwrap import dedent

from pex.common import open_zip, temporary_dir
from pex.interpreter import spawn_python_job
from pex.testing import WheelBuilder, make_project, temporary_content


def pex_project_dir():
  return subprocess.check_output(['git', 'rev-parse', '--show-toplevel']).decode('utf-8').strip()


@contextmanager
def bdist_pex(project_dir, bdist_args=None):
  with temporary_dir() as dist_dir:
    cmd = [
      'setup.py',
      '--command-packages', 'pex.commands',
      'bdist_pex', '--bdist-dir={}'.format(dist_dir)
    ]
    if bdist_args:
      cmd.extend(bdist_args)

    spawn_python_job(args=cmd, cwd=project_dir, pythonpath=[pex_project_dir()]).wait()
    dists = os.listdir(dist_dir)
    assert len(dists) == 1
    yield os.path.join(dist_dir, dists[0])


def assert_entry_points(entry_points):
  with make_project(name='my_app', entry_points=entry_points) as project_dir:
    with bdist_pex(project_dir) as my_app_pex:
      process = subprocess.Popen([my_app_pex], stdout=subprocess.PIPE)
      stdout, _ = process.communicate()
      assert '{pex_root}' not in os.listdir(project_dir)
      assert 0 == process.returncode
      assert stdout == b'hello world!\n'


def assert_pex_args_shebang(shebang):
  with make_project() as project_dir:
    pex_args = '--pex-args=--python-shebang="{}"'.format(shebang)
    with bdist_pex(project_dir, bdist_args=[pex_args]) as my_app_pex:
      with open(my_app_pex, 'rb') as fp:
        assert fp.readline().decode().rstrip() == shebang


def test_entry_points_dict():
  assert_entry_points({'console_scripts': ['my_app = my_app.my_module:do_something']})


def test_entry_points_ini_string():
  assert_entry_points(dedent("""
      [console_scripts]
      my_app=my_app.my_module:do_something
    """))


def test_pex_args_shebang_with_spaces():
  assert_pex_args_shebang('#!/usr/bin/env python')


def test_pex_args_shebang_without_spaces():
  assert_pex_args_shebang('#!/usr/bin/python')


def test_unwriteable_contents():
  my_app_setup_py = dedent("""
      from setuptools import setup

      setup(
        name='my_app',
        version='0.0.0',
        zip_safe=True,
        packages=['my_app'],
        include_package_data=True,
        package_data={'my_app': ['unwriteable.so']},
      )
    """)

  UNWRITEABLE_PERMS = 0o400
  with temporary_content({'setup.py': my_app_setup_py,
                          'my_app/__init__.py': '',
                          'my_app/unwriteable.so': 'so contents'},
                         perms=UNWRITEABLE_PERMS) as my_app_project_dir:
    my_app_whl = WheelBuilder(my_app_project_dir).bdist()

    with make_project(name='uses_my_app', install_reqs=['my_app']) as uses_my_app_project_dir:
      pex_args = '--pex-args=--disable-cache --no-pypi -f {}'.format(os.path.dirname(my_app_whl))
      with bdist_pex(uses_my_app_project_dir, bdist_args=[pex_args]) as uses_my_app_pex:
        with open_zip(uses_my_app_pex) as zf:
          unwriteable_sos = [path for path in zf.namelist()
                             if path.endswith('my_app/unwriteable.so')]
          assert 1 == len(unwriteable_sos)
          unwriteable_so = unwriteable_sos.pop()
          zf.extract(unwriteable_so, path=uses_my_app_project_dir)
          extract_dest = os.path.join(uses_my_app_project_dir, unwriteable_so)
          with open(extract_dest) as fp:
            assert 'so contents' == fp.read()



#!/usr/bin/env python3

import subprocess
import sys
from pathlib import Path

import pytoml as toml

PROJECT_METADATA = Path('pyproject.toml')


def python_requires() -> str:
  project_metadata = toml.loads(PROJECT_METADATA.read_text())
  return project_metadata['tool']['flit']['metadata']['requires-python'].strip()


def build_pex_pex() -> None:
  # NB: We do not include the subprocess extra (which would be spelled: `.[subprocess]`) since we
  # would then produce a pex that would not be consumable by all python interpreters otherwise
  # meeting `python_requires`; ie: we'd need to then come up with a deploy environment / deploy
  # tooling, that built subprocess32 for linux cp27m, cp27mu, pypy, ... etc. Even with all the work
  # expended to do this, we'd still miss some platform someone wanted to run the Pex PEX on. As
  # such, we just ship unadorned Pex which is pure-python and universal. Any user wanting the extra
  # is encouraged to build a Pex PEX for their particular platform themselves.
  pex_requirement = '.'

  args = [
    sys.executable,
    '-m', 'pex',
    '-v',
    '--disable-cache',
    '--no-build',
    '--no-compile',
    '--no-use-system-time',
    '--interpreter-constraint', python_requires(),
    '--python-shebang', '/usr/bin/env python',
    '--no-strip-pex-env',
    '--unzip',
    '-o', 'dist/pex',
    '-c', 'pex',
    pex_requirement
  ]
  subprocess.run(args, check=True)


if __name__ == '__main__':
  if not PROJECT_METADATA.is_file():
    print('This script must be run from the root of the Pex repo.', file=sys.stderr)
    sys.exit(1)

  build_pex_pex()



# coding=utf-8
# Copyright 2018 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os


class Bootstrap(object):
  """Supports introspection of the PEX bootstrap code."""

  _INSTANCE = None

  @classmethod
  def locate(cls):
    """Locates the active PEX bootstrap.

    :rtype: :class:`Bootstrap`
    """
    if cls._INSTANCE is None:
      bootstrap_path = __file__
      module_import_path = __name__.split('.')

      # For example, our __file__ might be requests.pex/.bootstrap/pex/bootstrap.pyc and our import
      # path pex.bootstrap; so we walk back through all the module components of our import path to
      # find the base sys.path entry where we were found (requests.pex/.bootstrap in this example).
      for _ in module_import_path:
        bootstrap_path = os.path.dirname(bootstrap_path)

      cls._INSTANCE = cls(sys_path_entry=bootstrap_path)
    return cls._INSTANCE

  def __init__(self, sys_path_entry):
    self._sys_path_entry = sys_path_entry
    self._realpath = os.path.realpath(self._sys_path_entry)

  def demote(self):
    """Demote the bootstrap code to the end of the `sys.path` so it is found last.

    :return: The list of un-imported bootstrap modules.
    :rtype: list of :class:`types.ModuleType`
    """
    import sys  # Grab a hold of `sys` early since we'll be un-importing our module in this process.

    unimported_modules = []
    for name, module in reversed(sorted(sys.modules.items())):
      if self.imported_from_bootstrap(module):
        unimported_modules.append(sys.modules.pop(name))

    sys.path[:] = [path for path in sys.path if os.path.realpath(path) != self._realpath]
    sys.path.append(self._sys_path_entry)

    return unimported_modules

  def imported_from_bootstrap(self, module):
    """Return ``True`` if the given ``module`` object was imported from bootstrap code.

    :param module: The module to check the provenance of.
    :type module: :class:`types.ModuleType`
    :rtype: bool
    """

    # A vendored module.
    path = getattr(module, '__file__', None)
    if path and os.path.realpath(path).startswith(self._realpath):
      return True

    # A vendored package.
    path = getattr(module, '__path__', None)
    if path and any(os.path.realpath(path_item).startswith(self._realpath)
                    for path_item in path):
      return True

    return False

  def __repr__(self):
    return '{cls}(sys_path_entry={sys_path_entry!r})'.format(cls=type(self).__name__,
                                                             sys_path_entry=self._sys_path_entry)



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
from collections import namedtuple


class DistributionScript(namedtuple('DistributionScript', ['dist', 'path'])):
  @classmethod
  def find(cls, dist, name):
    script_path = os.path.join(dist.location, 'bin', name)
    return cls(dist=dist, path=script_path) if os.path.isfile(script_path) else None

  def read_contents(self):
    with open(self.path) as fp:
      return fp.read()


def get_script_from_distributions(name, dists):
  for dist in dists:
    distribution_script = DistributionScript.find(dist, name)
    if distribution_script:
      return distribution_script


def get_entry_point_from_console_script(script, dists):
  # Check all distributions for the console_script "script". De-dup by dist key to allow for a
  # duplicate console script IFF the distribution is platform-specific and this is a multi-platform
  # pex.
  def get_entrypoint(dist):
    script_entry = dist.get_entry_map().get('console_scripts', {}).get(script)
    if script_entry is not None:
      # Entry points are of the form 'foo = bar', we just want the 'bar' part.
      return str(script_entry).split('=')[1].strip()

  entries = {}
  for dist in dists:
    entry_point = get_entrypoint(dist)
    if entry_point is not None:
      entries[dist.key] = (dist, entry_point)

  if len(entries) > 1:
    raise RuntimeError(
        'Ambiguous script specification %s matches multiple entry points:\n\t%s' % (
            script,
            '\n\t'.join('%r from %r' % (entry_point, dist)
                        for dist, entry_point in entries.values())))

  dist, entry_point = None, None
  if entries:
    dist, entry_point = next(iter(entries.values()))
  return dist, entry_point



# Copyright 2017 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import pytest

from pex.platforms import Platform

EXPECTED_BASE = [('py27', 'none', 'any'), ('py2', 'none', 'any')]


def test_platform():
  assert Platform('linux-x86_64', 'cp', '27', 'mu') == ('linux_x86_64', 'cp', '27', 'cp27mu')
  assert str(
    Platform('linux-x86_64', 'cp', '27', 'm')
  ) == 'linux_x86_64-cp-27-cp27m'


def test_platform_create():
  assert Platform.create('linux-x86_64-cp-27-cp27mu') == ('linux_x86_64', 'cp', '27', 'cp27mu')
  assert Platform.create('linux-x86_64-cp-27-mu') == ('linux_x86_64', 'cp', '27', 'cp27mu')
  assert Platform.create(
    'macosx-10.4-x86_64-cp-27-m') == ('macosx_10_4_x86_64', 'cp', '27', 'cp27m')


def test_platform_create_bad_platform_missing_fields():
  with pytest.raises(Platform.InvalidPlatformError):
    Platform.create('linux-x86_64')


def test_platform_create_bad_platform_empty_fields():
  with pytest.raises(Platform.InvalidPlatformError):
    Platform.create('linux-x86_64-cp--cp27mu')


def test_platform_create_noop():
  existing = Platform.create('linux-x86_64-cp-27-mu')
  assert Platform.create(existing) == existing



# Copyright 2019 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import errno
from threading import BoundedSemaphore, Event, Thread

from pex.compatibility import Queue, cpu_count
from pex.tracer import TRACER


class Job(object):
  """Represents a job spawned as a subprocess.

  Presents a similar API to `subprocess` except where noted.
  """

  class Error(Exception):
    """Indicates that a Job exited non-zero."""

  def __init__(self, command, process):
    """
    :param command: The command used to spawn the job process.
    :type command: list of str
    :param process: The spawned process handle.
    :type process: :class:`subprocess.Popen`
    """
    self._command = command
    self._process = process

  def wait(self):
    """Waits for the job to complete.

    :raises: :class:`Job.Error` if the job exited non-zero.
    """
    self._process.wait()
    self._check_returncode()

  def communicate(self, input=None):
    """Communicates with the job sending any input data to stdin and collecting stdout and stderr.

    :param input: Data to send to stdin of the job as per the `subprocess` API.
    :return: A tuple of the job's stdout and stderr as per the `subprocess` API.
    :raises: :class:`Job.Error` if the job exited non-zero.
    """
    stdout, stderr = self._process.communicate(input=input)
    self._check_returncode(stderr)
    return stdout, stderr

  def kill(self):
    """Terminates the job if it is still running.

    N.B.: This method is idempotent.
    """
    try:
      self._process.kill()
    except OSError as e:
      if e.errno != errno.ESRCH:
        raise e

  def _check_returncode(self, stderr=None):
    if self._process.returncode != 0:
      msg = 'Executing {} failed with {}'.format(' '.join(self._command), self._process.returncode)
      if stderr:
        msg += '\nSTDERR:\n{}'.format(stderr.decode('utf-8'))
      raise self.Error(msg)

  def __str__(self):
    return 'pid: {pid} -> {command}'.format(pid=self._process.pid, command=' '.join(self._command))


class SpawnedJob(object):
  """A handle to a spawned :class:`Job` and its associated result."""

  @classmethod
  def completed(cls, result):
    """Wrap an already completed result in a SpawnedJob.

    The returned job will no-op when `kill` is called since the job is already completed.

    :param result: The completed result.
    :return: A spawned job whose result is already complete.
    :rtype: :class:`SpawnedJob`
    """
    class Completed(SpawnedJob):
      def __init__(self):
        super(Completed, self).__init__(job=None, result_func=lambda: result)

      def kill(self):
        pass

      def __str__(self):
        return 'SpawnedJob.completed({})'.format(result)

    return Completed()

  @classmethod
  def wait(cls, job, result):
    """Wait for the job to complete and return a fixed result upon success.

    :param job: The spawned job.
    :type job: :class:`Job`
    :param result: The fixed success result.
    :return: A spawned job whose result is a side effect of the job (a written file, a populated
             directory, etc.).
    :rtype: :class:`SpawnedJob`
    """
    def wait_result_func():
      job.wait()
      return result

    return cls(job=job, result_func=wait_result_func)

  @classmethod
  def stdout(cls, job, result_func, input=None):
    """Wait for the job to complete and return a result derived from its stdout.

    :param job: The spawned job.
    :type job: :class:`Job`
    :param result_func: A function taking the stdout byte string collected from the spawned job and
                        returning the desired result.
    :param input: Optional input stream data to pass to the process as per the
                  `subprocess.Popen.communicate` API.
    :return: A spawned job whose result is derived from stdout contents.
    :rtype: :class:`SpawnedJob`
    """
    def stdout_result_func():
      stdout, _ = job.communicate(input=input)
      return result_func(stdout)

    return cls(job=job, result_func=stdout_result_func)

  def __init__(self, job, result_func):
    """Not intended for direct use, see `wait` and `stdout` factories."""
    self._job = job
    self._result_func = result_func

  def await_result(self):
    """Waits for the spawned job to complete and returns its result."""
    return self._result_func()

  def kill(self):
    """Terminates the spawned job if it's not already complete."""
    self._job.kill()

  def __str__(self):
    return str(self._job)


_CPU_COUNT = cpu_count()
_ABSOLUTE_MAX_JOBS = _CPU_COUNT * 2


DEFAULT_MAX_JOBS = _CPU_COUNT
"""The default maximum number of parallel jobs PEX should use."""


def _sanitize_max_jobs(max_jobs=None):
  assert max_jobs is None or isinstance(max_jobs, int)
  if max_jobs is None or max_jobs <= 0:
    return DEFAULT_MAX_JOBS
  else:
    return min(max_jobs, _ABSOLUTE_MAX_JOBS)


def execute_parallel(inputs, spawn_func, raise_type=None, max_jobs=None):
  """Execute jobs for the given inputs in parallel.

  :param int max_jobs: The maximum number of parallel jobs to spawn.
  :param inputs: An iterable of the data to parallelize over `spawn_func`.
  :param spawn_func: A function taking a single input and returning a :class:`SpawnedJob`.
  :param raise_type: An optional type that takes a single string argument and will be used to
                     construct a raiseable value when any of the spawned jobs errors. `None` by
                     default which indicates spawned jobs that error should be skipped; i.e.: the
                     returned iterator over spawned job results will return less results than
                     inputs.
  :returns: An iterator over the spawned job results as they come in.
  :raises: A `raise_type` exception if any individual job errors and `raise_type` is not `None`.
  """
  size = _sanitize_max_jobs(max_jobs)
  TRACER.log('Spawning a maximum of {} parallel jobs to process:\n  {}'
             .format(size, '\n  '.join(map(str, inputs))),
             V=9)

  class SpawnError(Exception):
    def __init__(self, item, error):
      self.item = item
      self.error = error

  stop = Event()  # Used as a signal to stop spawning further jobs once any one job fails.
  job_slots = BoundedSemaphore(value=size)
  done_sentinel = object()
  spawned_job_queue = Queue()  # Queue[Union[SpawnedJob, SpawnError, Literal[done_sentinel]]]

  def spawn_jobs():
    for item in inputs:
      if stop.is_set():
        break
      job_slots.acquire()
      try:
        result = spawn_func(item)
      except Exception as e:
        result = SpawnError(item, e)
      finally:
        spawned_job_queue.put(result)
    spawned_job_queue.put(done_sentinel)

  spawner = Thread(name='PEX Parallel Job Spawner', target=spawn_jobs)
  spawner.daemon = True
  spawner.start()

  error = None
  while True:
    item = spawned_job_queue.get()

    if item is done_sentinel:
      if error:
        raise error
      return

    try:
      if isinstance(item, SpawnError):
        if raise_type:
          error = item.error
        else:
          # Otherwise, if we were given no `raise_type`, we skip over the input that raised.
          TRACER.log('Failed to spawn a job for {}: {} raised {}'
                     .format(item.item, spawn_func, item.error))
      elif error is not None:  # I.E.: `item` is not an exception, but there was a prior exception.
        item.kill()
      else:
        try:
          yield item.await_result()
        except Job.Error as e:
          if raise_type:
            # Fail fast and proceed to kill all outstanding spawned jobs.
            stop.set()
            error = raise_type('{} raised {}'.format(item, e))
          else:
            # Otherwise, if we were given no `raise_type`, we continue on and just log the failure.
            TRACER.log('{} raised {}'.format(item, e))
    finally:
      job_slots.release()



# Copyright 2016 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import contextlib
import errno
import os
from contextlib import contextmanager

import pytest

from pex.common import (
    AtomicDirectory,
    Chroot,
    PermPreservingZipFile,
    atomic_directory,
    can_write_dir,
    chmod_plus_x,
    temporary_dir,
    touch
)

try:
  from unittest import mock
except ImportError:
  import mock


@contextmanager
def maybe_raises(exception=None):
  @contextmanager
  def noop():
    yield

  with (noop() if exception is None else pytest.raises(exception)):
    yield


def atomic_directory_finalize_test(errno, expect_raises=None):
  with mock.patch('os.rename', spec_set=True, autospec=True) as mock_rename:
    mock_rename.side_effect = OSError(errno, os.strerror(errno))
    with maybe_raises(expect_raises):
      AtomicDirectory('to.dir').finalize()


def test_atomic_directory_finalize_eexist():
  atomic_directory_finalize_test(errno.EEXIST)


def test_atomic_directory_finalize_enotempty():
  atomic_directory_finalize_test(errno.ENOTEMPTY)


def test_atomic_directory_finalize_eperm():
  atomic_directory_finalize_test(errno.EPERM, expect_raises=OSError)


def test_atomic_directory_empty_workdir_finalize():
  with temporary_dir() as sandbox:
    target_dir = os.path.join(sandbox, 'target_dir')
    assert not os.path.exists(target_dir)

    with atomic_directory(target_dir) as work_dir:
      assert work_dir is not None
      assert os.path.exists(work_dir)
      assert os.path.isdir(work_dir)
      assert [] == os.listdir(work_dir)

      touch(os.path.join(work_dir, 'created'))

      assert not os.path.exists(target_dir)

    assert not os.path.exists(work_dir), 'The work_dir should always be cleaned up.'
    assert os.path.exists(os.path.join(target_dir, 'created'))


def test_atomic_directory_empty_workdir_failure():
  class SimulatedRuntimeError(RuntimeError):
    pass

  with temporary_dir() as sandbox:
    target_dir = os.path.join(sandbox, 'target_dir')
    with pytest.raises(SimulatedRuntimeError):
      with atomic_directory(target_dir) as work_dir:
        touch(os.path.join(work_dir, 'created'))
        raise SimulatedRuntimeError()

    assert not os.path.exists(work_dir), 'The work_dir should always be cleaned up.'
    assert not os.path.exists(target_dir), (
      'When the context raises the work_dir it was given should not be moved to the target_dir.'
    )


def test_atomic_directory_empty_workdir_finalized():
  with temporary_dir() as target_dir:
    with atomic_directory(target_dir) as work_dir:
      assert work_dir is None, 'When the target_dir exists no work_dir should be created.'


def extract_perms(path):
  return oct(os.stat(path).st_mode)


@contextlib.contextmanager
def zip_fixture():
  with temporary_dir() as target_dir:
    one = os.path.join(target_dir, 'one')
    touch(one)

    two = os.path.join(target_dir, 'two')
    touch(two)
    chmod_plus_x(two)

    assert extract_perms(one) != extract_perms(two)

    zip_file = os.path.join(target_dir, 'test.zip')
    with contextlib.closing(PermPreservingZipFile(zip_file, 'w')) as zf:
      zf.write(one, 'one')
      zf.write(two, 'two')

    yield zip_file, os.path.join(target_dir, 'extract'), one, two


def test_perm_preserving_zipfile_extractall():
  with zip_fixture() as (zip_file, extract_dir, one, two):
    with contextlib.closing(PermPreservingZipFile(zip_file)) as zf:
      zf.extractall(extract_dir)

      assert extract_perms(one) == extract_perms(os.path.join(extract_dir, 'one'))
      assert extract_perms(two) == extract_perms(os.path.join(extract_dir, 'two'))


def test_perm_preserving_zipfile_extract():
  with zip_fixture() as (zip_file, extract_dir, one, two):
    with contextlib.closing(PermPreservingZipFile(zip_file)) as zf:
      zf.extract('one', path=extract_dir)
      zf.extract('two', path=extract_dir)

      assert extract_perms(one) == extract_perms(os.path.join(extract_dir, 'one'))
      assert extract_perms(two) == extract_perms(os.path.join(extract_dir, 'two'))


def assert_chroot_perms(copyfn):
  with temporary_dir() as src:
    one = os.path.join(src, 'one')
    touch(one)

    two = os.path.join(src, 'two')
    touch(two)
    chmod_plus_x(two)

    with temporary_dir() as dst:
      chroot = Chroot(dst)
      copyfn(chroot, one, 'one')
      copyfn(chroot, two, 'two')
      assert extract_perms(one) == extract_perms(os.path.join(chroot.path(), 'one'))
      assert extract_perms(two) == extract_perms(os.path.join(chroot.path(), 'two'))

      zip_path = os.path.join(src, 'chroot.zip')
      chroot.zip(zip_path)
      with temporary_dir() as extract_dir:
        with contextlib.closing(PermPreservingZipFile(zip_path)) as zf:
          zf.extractall(extract_dir)

          assert extract_perms(one) == extract_perms(os.path.join(extract_dir, 'one'))
          assert extract_perms(two) == extract_perms(os.path.join(extract_dir, 'two'))


def test_chroot_perms_copy():
  assert_chroot_perms(Chroot.copy)


def test_chroot_perms_link_same_device():
  assert_chroot_perms(Chroot.link)


def test_chroot_perms_link_cross_device():
  with mock.patch('os.link', spec_set=True, autospec=True) as mock_link:
    expected_errno = errno.EXDEV
    mock_link.side_effect = OSError(expected_errno, os.strerror(expected_errno))

    assert_chroot_perms(Chroot.link)


def test_can_write_dir_writeable_perms():
  with temporary_dir() as writeable:
    assert can_write_dir(writeable)

    path = os.path.join(writeable, 'does/not/exist/yet')
    assert can_write_dir(path)
    touch(path)
    assert not can_write_dir(path), 'Should not be able to write to a file.'


def test_can_write_dir_unwriteable_perms():
  with temporary_dir() as writeable:
    no_perms_path = os.path.join(writeable, 'no_perms')
    os.mkdir(no_perms_path, 0o444)
    assert not can_write_dir(no_perms_path)

    path_that_does_not_exist_yet = os.path.join(no_perms_path, 'does/not/exist/yet')
    assert not can_write_dir(path_that_does_not_exist_yet)

    os.chmod(no_perms_path, 0o744)
    assert can_write_dir(no_perms_path)
    assert can_write_dir(path_that_does_not_exist_yet)



# Copyright 2017 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from collections import namedtuple
from textwrap import dedent

from pex.interpreter import PythonInterpreter


class Platform(namedtuple('Platform', ['platform', 'impl', 'version', 'abi'])):
  """Represents a target platform and it's extended interpreter compatibility
  tags (e.g. implementation, version and ABI)."""

  class InvalidPlatformError(Exception):
    """Indicates an invalid platform string."""

  SEP = '-'

  @classmethod
  def create(cls, platform):
    if isinstance(platform, Platform):
      return platform

    platform = platform.lower()
    try:
      platform, impl, version, abi = platform.rsplit(cls.SEP, 3)
      return cls(platform, impl, version, abi)
    except ValueError:
      raise cls.InvalidPlatformError(dedent("""\
        Not a valid platform specifier: {}

        Platform strings must be in one of two forms:
        1. Canonical: <platform>-<python impl abbr>-<python version>-<abi>
        2. Abbreviated: <platform>-<python impl abbr>-<python version>-<abbr abi>

        Given a canonical platform string for CPython 3.7.5 running on 64 bit linux of:
          linux-x86_64-cp-37-cp37m

        Where the fields above are:
        + <platform>: linux-x86_64 
        + <python impl abbr>: cp
        + <python version>: 37
        + <abi>: cp37m

        The abbreviated platform string is:
          linux-x86_64-cp-37-m

        These fields stem from wheel name conventions as outlined in
        https://www.python.org/dev/peps/pep-0427#file-name-convention and influenced by
        https://www.python.org/dev/peps/pep-0425.
        """.format(platform)))

  @staticmethod
  def _maybe_prefix_abi(impl, version, abi):
    if impl != 'cp':
      return abi
    # N.B. This permits CPython users to pass in simpler extended platform
    # strings like `linux-x86_64-cp-27-mu` vs e.g. `linux-x86_64-cp-27-cp27mu`.
    impl_ver = ''.join((impl, version))
    return abi if abi.startswith(impl_ver) else ''.join((impl_ver, abi))

  def __new__(cls, platform, impl, version, abi):
    if not all((platform, impl, version, abi)):
      raise cls.InvalidPlatformError(
        'Platform specifiers cannot have blank fields. Given platform={platform!r}, impl={impl!r}, '
        'version={version!r}, abi={abi!r}'.format(
          platform=platform, impl=impl, version=version, abi=abi
        )
      )
    platform = platform.replace('-', '_').replace('.', '_')
    abi = cls._maybe_prefix_abi(impl, version, abi)
    return super(Platform, cls).__new__(cls, platform, impl, version, abi)

  @classmethod
  def of_interpreter(cls, intepreter=None):
    intepreter = intepreter or PythonInterpreter.get()
    identity = intepreter.identity
    impl, version = identity.python_tag[:2], identity.python_tag[2:]
    return cls(platform=identity.platform_tag,
               impl=impl,
               version=version,
               abi=identity.abi_tag)

  @classmethod
  def current(cls):
    return cls.of_interpreter()

  def __str__(self):
    return self.SEP.join(self)



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import contextlib
import os
import platform
import random
import subprocess
import sys
from collections import namedtuple
from contextlib import contextmanager
from textwrap import dedent

from pex.common import open_zip, safe_mkdir, safe_mkdtemp, safe_rmtree, temporary_dir, touch
from pex.distribution_target import DistributionTarget
from pex.executor import Executor
from pex.interpreter import PythonInterpreter
from pex.pex import PEX
from pex.pex_builder import PEXBuilder
from pex.pip import get_pip
from pex.third_party.pkg_resources import Distribution
from pex.util import DistributionHelper, named_temporary_file

PY_VER = sys.version_info[:2]
IS_PYPY = hasattr(sys, 'pypy_version_info')
NOT_CPYTHON27 = IS_PYPY or PY_VER != (2, 7)
NOT_CPYTHON36 = IS_PYPY or PY_VER != (3, 6)
IS_LINUX = platform.system() == 'Linux'
IS_NOT_LINUX = not IS_LINUX
NOT_CPYTHON27_OR_OSX = NOT_CPYTHON27 or IS_NOT_LINUX
NOT_CPYTHON36_OR_LINUX = NOT_CPYTHON36 or IS_LINUX


@contextlib.contextmanager
def temporary_filename():
  """Creates a temporary filename.

  This is useful when you need to pass a filename to an API. Windows requires all
  handles to a file be closed before deleting/renaming it, so this makes it a bit
  simpler."""
  with named_temporary_file() as fp:
    fp.write(b'')
    fp.close()
    yield fp.name


def random_bytes(length):
  return ''.join(
      map(chr, (random.randint(ord('a'), ord('z')) for _ in range(length)))).encode('utf-8')


def get_dep_dist_names_from_pex(pex_path, match_prefix=''):
  """Given an on-disk pex, extract all of the unique first-level paths under `.deps`."""
  with open_zip(pex_path) as pex_zip:
    dep_gen = (f.split(os.sep)[1] for f in pex_zip.namelist() if f.startswith('.deps/'))
    return set(item for item in dep_gen if item.startswith(match_prefix))


@contextlib.contextmanager
def temporary_content(content_map, interp=None, seed=31337, perms=0o644):
  """Write content to disk where content is map from string => (int, string).

     If target is int, write int random bytes.  Otherwise write contents of string."""
  random.seed(seed)
  interp = interp or {}
  with temporary_dir() as td:
    for filename, size_or_content in content_map.items():
      dest = os.path.join(td, filename)
      safe_mkdir(os.path.dirname(dest))
      with open(dest, 'wb') as fp:
        if isinstance(size_or_content, int):
          fp.write(random_bytes(size_or_content))
        else:
          fp.write((size_or_content % interp).encode('utf-8'))
      os.chmod(dest, perms)
    yield td


def yield_files(directory):
  for root, _, files in os.walk(directory):
    for f in files:
      filename = os.path.join(root, f)
      rel_filename = os.path.relpath(filename, directory)
      yield filename, rel_filename


def write_zipfile(directory, dest, reverse=False):
  with open_zip(dest, 'w') as zf:
    for filename, rel_filename in sorted(yield_files(directory), reverse=reverse):
      zf.write(filename, arcname=rel_filename)
  return dest


@contextlib.contextmanager
def make_project(name='my_project',
                 version='0.0.0',
                 zip_safe=True,
                 install_reqs=None,
                 extras_require=None,
                 entry_points=None):

  project_content = {
    'setup.py': dedent('''
      from setuptools import setup

      setup(
        name=%(project_name)r,
        version=%(version)r,
        zip_safe=%(zip_safe)r,
        packages=[%(project_name)r],
        scripts=[
          'scripts/hello_world',
          'scripts/shell_script',
        ],
        package_data={%(project_name)r: ['package_data/*.dat']},
        install_requires=%(install_requires)r,
        extras_require=%(extras_require)r,
        entry_points=%(entry_points)r,
      )
    '''),
    'scripts/hello_world': '#!/usr/bin/env python\nprint("hello world!")\n',
    'scripts/shell_script': '#!/usr/bin/env bash\necho hello world\n',
    os.path.join(name, '__init__.py'): 0,
    os.path.join(name, 'my_module.py'): 'def do_something():\n  print("hello world!")\n',
    os.path.join(name, 'package_data/resource1.dat'): 1000,
    os.path.join(name, 'package_data/resource2.dat'): 1000,
  }

  interp = {'project_name': name,
            'version': version,
            'zip_safe': zip_safe,
            'install_requires': install_reqs or [],
            'extras_require': extras_require or {},
            'entry_points': entry_points or {}}

  with temporary_content(project_content, interp=interp) as td:
    yield td


class WheelBuilder(object):
  """Create a wheel distribution from an unpacked setup.py-based project."""

  class BuildFailure(Exception):
    pass

  def __init__(self, source_dir, interpreter=None, wheel_dir=None):
    """Create a wheel from an unpacked source distribution in source_dir."""
    self._source_dir = source_dir
    self._wheel_dir = wheel_dir or safe_mkdtemp()
    self._interpreter = interpreter or PythonInterpreter.get()

  def bdist(self):
    get_pip().spawn_build_wheels(
      distributions=[self._source_dir],
      wheel_dir=self._wheel_dir,
      interpreter=self._interpreter
    ).wait()
    dists = os.listdir(self._wheel_dir)
    if len(dists) == 0:
      raise self.BuildFailure('No distributions were produced!')
    elif len(dists) > 1:
      raise self.BuildFailure('Ambiguous source distributions found: %s' % (' '.join(dists)))
    else:
      return os.path.join(self._wheel_dir, dists[0])


@contextlib.contextmanager
def built_wheel(name='my_project',
                version='0.0.0',
                zip_safe=True,
                install_reqs=None,
                extras_require=None,
                interpreter=None,
                **kwargs):

  with make_project(name=name,
                    version=version,
                    zip_safe=zip_safe,
                    install_reqs=install_reqs,
                    extras_require=extras_require) as td:
    builder = WheelBuilder(td, interpreter=interpreter, **kwargs)
    yield builder.bdist()


@contextlib.contextmanager
def make_source_dir(name='my_project', version='0.0.0', install_reqs=None, extras_require=None):
  with make_project(name=name,
                    version=version,
                    install_reqs=install_reqs,
                    extras_require=extras_require) as td:
    yield td


@contextlib.contextmanager
def make_bdist(name='my_project', version='0.0.0', zip_safe=True, interpreter=None, **kwargs):
  with built_wheel(name=name,
                   version=version,
                   zip_safe=zip_safe,
                   interpreter=interpreter,
                   **kwargs) as dist_location:

    install_dir = os.path.join(safe_mkdtemp(), os.path.basename(dist_location))
    get_pip().spawn_install_wheel(
      wheel=dist_location,
      install_dir=install_dir,
      target=DistributionTarget.for_interpreter(interpreter)
    ).wait()
    yield DistributionHelper.distribution_from_path(install_dir)


COVERAGE_PREAMBLE = """
try:
  from coverage import coverage
  cov = coverage(auto_data=True, data_suffix=True)
  cov.start()
except ImportError:
  pass
"""


def write_simple_pex(td,
                     exe_contents=None,
                     dists=None,
                     sources=None,
                     coverage=False,
                     interpreter=None,
                     pex_info=None):
  """Write a pex file that optionally contains an executable entry point.

  :param str td: temporary directory path
  :param str exe_contents: entry point python file
  :param dists: distributions to include, typically sdists or bdists
  :type: list of :class:`pex.third_party.pkg_resources.Distribution`
  :param sources: sources to include, as a list of pairs (env_filename, contents)
  :type sources: list of (str, str)
  :param bool coverage: include coverage header
  :param interpreter: a custom interpreter to use to build the pex
  :type interpreter: :class:`pex.interpreter.PythonInterpreter`
  :param pex_info: a custom PexInfo to use to build the pex.
  :type pex_info: :class:`pex.pex_info.PexInfo`
  """
  dists = dists or []
  sources = sources or []

  safe_mkdir(td)

  pb = PEXBuilder(path=td,
                  preamble=COVERAGE_PREAMBLE if coverage else None,
                  interpreter=interpreter,
                  pex_info=pex_info)

  for dist in dists:
    pb.add_dist_location(dist.location if isinstance(dist, Distribution) else dist)

  for env_filename, contents in sources:
    src_path = os.path.join(td, env_filename)
    safe_mkdir(os.path.dirname(src_path))
    with open(src_path, 'w') as fp:
      fp.write(contents)
    pb.add_source(src_path, env_filename)

  if exe_contents:
    with open(os.path.join(td, 'exe.py'), 'w') as fp:
      fp.write(exe_contents)
    pb.set_executable(os.path.join(td, 'exe.py'))

  pb.freeze()

  return pb


class IntegResults(namedtuple('results', ['output', 'error', 'return_code'])):
  """Convenience object to return integration run results."""

  def assert_success(self):
    assert self.return_code == 0, ('integration test failed: return_code={}, output={}, error={}'
                                   .format(self.return_code, self.output, self.error))

  def assert_failure(self):
    assert self.return_code != 0


def run_pex_command(args, env=None, python=None, quiet=False):
  """Simulate running pex command for integration testing.

  This is different from run_simple_pex in that it calls the pex command rather
  than running a generated pex.  This is useful for testing end to end runs
  with specific command line arguments or env options.
  """
  cmd = [python or sys.executable, '-mpex']
  if not quiet:
    cmd.append('-vvvvv')
  cmd.extend(args)
  process = Executor.open_process(cmd=cmd, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
  output, error = process.communicate()
  return IntegResults(output.decode('utf-8'), error.decode('utf-8'), process.returncode)


def run_simple_pex(pex, args=(), interpreter=None, stdin=None, **kwargs):
  p = PEX(pex, interpreter=interpreter)
  process = p.run(args=args,
                  blocking=False,
                  stdin=subprocess.PIPE,
                  stdout=subprocess.PIPE,
                  stderr=subprocess.STDOUT,
                  **kwargs)
  stdout, _ = process.communicate(input=stdin)
  return stdout.replace(b'\r', b''), process.returncode


def run_simple_pex_test(body, args=(), env=None, dists=None, coverage=False, interpreter=None):
  with temporary_dir() as td1, temporary_dir() as td2:
    pb = write_simple_pex(td1, body, dists=dists, coverage=coverage, interpreter=interpreter)
    pex = os.path.join(td2, 'app.pex')
    pb.build(pex)
    return run_simple_pex(pex, args=args, env=env, interpreter=interpreter)


def bootstrap_python_installer(dest):
  safe_rmtree(dest)
  for _ in range(3):
    try:
      subprocess.check_call(
        ['git', 'clone', 'https://github.com/pyenv/pyenv.git', dest]
      )
    except subprocess.CalledProcessError as e:
      print('caught exception: %r' % e)
      continue
    else:
      break
  else:
    raise RuntimeError("Helper method could not clone pyenv from git after 3 tries")
  # Create an empty file indicating the fingerprint of the correct set of test interpreters.
  touch(os.path.join(dest, _INTERPRETER_SET_FINGERPRINT))


# NB: We keep the pool of bootstrapped interpreters as small as possible to avoid timeouts in CI
# otherwise encountered when fetching and building too many on a cache miss. In the past we had
# issues with the combination of 7 total unique interpreter versions and a Travis-CI timeout of 50
# minutes for a shard.
PY27 = '2.7.15'
PY35 = '3.5.6'
PY36 = '3.6.6'

_VERSIONS = (PY27, PY35, PY36)
# This is the filename of a sentinel file that sits in the pyenv root directory.
# Its purpose is to indicate whether pyenv has the correct interpreters installed
# and will be useful for indicating whether we should trigger a reclone to update
# pyenv.
_INTERPRETER_SET_FINGERPRINT = '_'.join(_VERSIONS) + '_pex_fingerprint'


def ensure_python_distribution(version):
  if version not in _VERSIONS:
    raise ValueError('Please constrain version to one of {}'.format(_VERSIONS))

  pyenv_root = os.path.join(os.getcwd(), '.pyenv_test')
  interpreter_location = os.path.join(pyenv_root, 'versions', version)
  pyenv = os.path.join(pyenv_root, 'bin', 'pyenv')
  pip = os.path.join(interpreter_location, 'bin', 'pip')

  if not os.path.exists(os.path.join(pyenv_root, _INTERPRETER_SET_FINGERPRINT)):
    bootstrap_python_installer(pyenv_root)

  if not os.path.exists(interpreter_location):
    env = os.environ.copy()
    env['PYENV_ROOT'] = pyenv_root
    if sys.platform.lower() == 'linux':
      env['CONFIGURE_OPTS'] = '--enable-shared'
    subprocess.check_call([pyenv, 'install', '--keep', version], env=env)
    subprocess.check_call([pip, 'install', '-U', 'pip'])

  python = os.path.join(interpreter_location, 'bin', 'python' + version[0:3])
  return python, pip


def ensure_python_interpreter(version):
  python, _ = ensure_python_distribution(version)
  return python


@contextmanager
def environment_as(**kwargs):
  existing = {key: os.environ.get(key) for key in kwargs}

  def adjust_environment(mapping):
    for key, value in mapping.items():
      if value is not None:
        os.environ[key] = value
      else:
        del os.environ[key]

  adjust_environment(kwargs)
  try:
    yield
  finally:
    adjust_environment(existing)



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
from contextlib import contextmanager
from optparse import OptionParser
from tempfile import NamedTemporaryFile

import pytest

from pex.bin.pex import build_pex, configure_clp, configure_clp_pex_resolution
from pex.common import safe_copy, temporary_dir
from pex.compatibility import nested, to_bytes
from pex.platforms import Platform
from pex.testing import (
    PY27,
    built_wheel,
    ensure_python_interpreter,
    run_pex_command,
    run_simple_pex
)


@contextmanager
def option_parser():
  yield OptionParser()


def test_clp_no_pypi_option():
  with option_parser() as parser:
    configure_clp_pex_resolution(parser)
    options, _ = parser.parse_args(args=[])
    assert len(options.indexes) == 1
    options, _ = parser.parse_args(args=['--no-pypi'])
    assert len(options.indexes) == 0, '--no-pypi should remove the pypi index.'


def test_clp_pypi_option_duplicate():
  with option_parser() as parser:
    configure_clp_pex_resolution(parser)
    options, _ = parser.parse_args(args=[])
    assert len(options.indexes) == 1
    options2, _ = parser.parse_args(args=['--pypi'])
    assert len(options2.indexes) == 1
    assert options.indexes == options2.indexes


def test_clp_find_links_option():
  with option_parser() as parser:
    configure_clp_pex_resolution(parser)
    options, _ = parser.parse_args(args=['-f', 'http://www.example.com'])
    assert len(options.indexes) == 1
    assert len(options.find_links) == 1


def test_clp_index_option():
  with option_parser() as parser:
    configure_clp_pex_resolution(parser)
    options, _ = parser.parse_args(args=[])
    assert len(options.indexes) == 1
    options2, _ = parser.parse_args(args=['-i', 'http://www.example.com'])
    assert len(options.indexes) == 2
    assert options2.indexes[0] == options.indexes[0]
    assert options2.indexes[1] == 'http://www.example.com'


def test_clp_index_option_render():
  with option_parser() as parser:
    configure_clp_pex_resolution(parser)
    options, _ = parser.parse_args(args=['--index', 'http://www.example.com'])
    assert ['https://pypi.org/simple', 'http://www.example.com'] == [
      str(idx) for idx in options.indexes
    ]


def test_clp_build_precedence():
  with option_parser() as parser:
    configure_clp_pex_resolution(parser)

    options, _ = parser.parse_args(args=['--no-build'])
    assert not options.build
    options, _ = parser.parse_args(args=['--build'])
    assert options.build

    options, _ = parser.parse_args(args=['--no-wheel'])
    assert not options.use_wheel

    options, _ = parser.parse_args(args=['--wheel'])
    assert options.use_wheel


# Make sure that we're doing append and not replace
def test_clp_requirements_txt():
  parser = configure_clp()
  options, _ = parser.parse_args(args='-r requirements1.txt -r requirements2.txt'.split())
  assert options.requirement_files == ['requirements1.txt', 'requirements2.txt']


def test_clp_constraints_txt():
  parser = configure_clp()
  options, _ = parser.parse_args(args='--constraint requirements1.txt'.split())
  assert options.constraint_files == ['requirements1.txt']


def test_clp_preamble_file():
  with NamedTemporaryFile() as tmpfile:
    tmpfile.write(to_bytes('print "foo!"'))
    tmpfile.flush()

    parser = configure_clp()
    options, reqs = parser.parse_args(args=['--preamble-file', tmpfile.name])
    assert options.preamble_file == tmpfile.name

    pex_builder = build_pex(reqs, options)
    assert pex_builder._preamble == 'print "foo!"'


def test_clp_prereleases():
  with option_parser() as parser:
    configure_clp_pex_resolution(parser)

    options, _ = parser.parse_args(args=[])
    assert not options.allow_prereleases

    options, _ = parser.parse_args(args=['--no-pre'])
    assert not options.allow_prereleases

    options, _ = parser.parse_args(args=['--pre'])
    assert options.allow_prereleases


def test_clp_prereleases_resolver():
  with nested(built_wheel(name='prerelease-dep', version='1.2.3b1'),
              built_wheel(name='transitive-dep', install_reqs=['prerelease-dep']),
              built_wheel(name='dep', install_reqs=['prerelease-dep>=1.2', 'transitive-dep']),
              temporary_dir(),
              temporary_dir()) as (prerelease_dep, transitive_dep, dep, dist_dir, cache_dir):

    for dist in (prerelease_dep, transitive_dep, dep):
      safe_copy(dist, os.path.join(dist_dir, os.path.basename(dist)))

    parser = configure_clp()

    options, reqs = parser.parse_args(args=[
      '--no-index',
      '--find-links', dist_dir,
      '--cache-dir', cache_dir,  # Avoid dangling {pex_root}.
      '--no-pre',
      'dep'
    ])
    assert not options.allow_prereleases

    with pytest.raises(SystemExit, message='Should have failed to resolve prerelease dep'):
      build_pex(reqs, options)

    # When we specify `--pre`, allow_prereleases is True
    options, reqs = parser.parse_args(args=[
      '--no-index',
      '--find-links', dist_dir,
      '--cache-dir', cache_dir,  # Avoid dangling {pex_root}.
      '--pre',
      'dep'
    ])
    assert options.allow_prereleases

    # Without a corresponding fix in pex.py, this test failed for a dependency requirement of
    # dep==1.2.3b1 from one package and just dep (any version accepted) from another package.
    # The failure was an exit from build_pex() with the message:
    #
    # Could not satisfy all requirements for dep==1.2.3b1:
    #     dep==1.2.3b1, dep
    #
    # With a correct behavior the assert line is reached and pex_builder object created.
    pex_builder = build_pex(reqs, options)
    assert pex_builder is not None
    assert len(pex_builder.info.distributions) == 3, 'Should have resolved deps'


def test_build_pex():
  with temporary_dir() as sandbox:
    pex_path = os.path.join(sandbox, 'pex')
    results = run_pex_command(['ansicolors==1.1.8', '--output-file', pex_path])
    results.assert_success()
    stdout, returncode = run_simple_pex(
      pex=pex_path,
      args=['-c', 'import colors; print(" ".join(colors.COLORS))']
    )
    assert 0 == returncode
    assert b'black red green yellow blue magenta cyan white' == stdout.strip()


def assert_run_pex(python=None, pex_args=None):
  pex_args = list(pex_args) if pex_args else []
  results = run_pex_command(
    python=python,
    args=pex_args + [
      'ansicolors==1.1.8', '--', '-c', 'import colors; print(" ".join(colors.COLORS))'
    ],
    quiet=True
  )
  results.assert_success()
  assert 'black red green yellow blue magenta cyan white' == results.output.strip()
  return results.error.splitlines()


def test_run_pex():
  incompatible_platforms_warning_msg = 'WARNING: attempting to run PEX with incompatible platforms!'

  assert incompatible_platforms_warning_msg not in assert_run_pex()
  assert incompatible_platforms_warning_msg not in assert_run_pex(pex_args=['--platform=current'])
  assert incompatible_platforms_warning_msg not in assert_run_pex(
    pex_args=['--platform={}'.format(Platform.current())]
  )

  py27 = ensure_python_interpreter(PY27)
  stderr_lines = assert_run_pex(python=py27, pex_args=['--platform=macosx-10.13-x86_64-cp-37-m'])
  assert incompatible_platforms_warning_msg in stderr_lines



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import importlib
import itertools
import os
import site
import sys
import zipfile
from collections import OrderedDict, defaultdict

from pex import dist_metadata, pex_builder, pex_warnings
from pex.bootstrap import Bootstrap
from pex.common import atomic_directory, die, open_zip
from pex.interpreter import PythonInterpreter
from pex.orderedset import OrderedSet
from pex.platforms import Platform
from pex.third_party.packaging import tags
from pex.third_party.pkg_resources import DistributionNotFound, Environment, Requirement, WorkingSet
from pex.tracer import TRACER
from pex.util import CacheHelper, DistributionHelper


def _import_pkg_resources():
  try:
    import pkg_resources  # vendor:skip
    return pkg_resources, False
  except ImportError:
    from pex import third_party
    third_party.install(expose=['setuptools'])
    import pkg_resources  # vendor:skip
    return pkg_resources, True


class PEXEnvironment(Environment):
  class _CachingZipImporter(object):
    class _CachingLoader(object):
      def __init__(self, delegate):
        self._delegate = delegate

      def load_module(self, fullname):
        loaded = sys.modules.get(fullname)
        # Technically a PEP-302 loader should re-load the existing module object here - notably
        # re-exec'ing the code found in the zip against the existing module __dict__. We don't do
        # this since the zip is assumed immutable during our run and this is enough to work around
        # the issue.
        if not loaded:
          loaded = self._delegate.load_module(fullname)
          loaded.__loader__ = self
        return loaded

    _REGISTERED = False

    @classmethod
    def _ensure_namespace_handler_registered(cls):
      if not cls._REGISTERED:
        pkg_resources, _ = _import_pkg_resources()
        pkg_resources.register_namespace_handler(cls, pkg_resources.file_ns_handler)
        cls._REGISTERED = True

    def __init__(self, path):
      import zipimport
      self._delegate = zipimport.zipimporter(path)

    def find_module(self, fullname, path=None):
      loader = self._delegate.find_module(fullname, path)
      if loader is None:
        return None
      self._ensure_namespace_handler_registered()
      caching_loader = self._CachingLoader(loader)
      return caching_loader

  @classmethod
  def _install_pypy_zipimporter_workaround(cls, pex_file):
    # The pypy zipimporter implementation always freshly loads a module instead of re-importing
    # when the module already exists in sys.modules. This breaks the PEP-302 importer protocol and
    # violates pkg_resources assumptions based on that protocol in its handling of namespace
    # packages. See: https://bitbucket.org/pypy/pypy/issues/1686

    def pypy_zipimporter_workaround(path):
      import os

      if not path.startswith(pex_file) or '.' in os.path.relpath(path, pex_file):
        # We only need to claim the pex zipfile root modules.
        #
        # The protocol is to raise if we don't want to hook the given path.
        # See: https://www.python.org/dev/peps/pep-0302/#specification-part-2-registering-hooks
        raise ImportError()

      return cls._CachingZipImporter(path)

    for path in list(sys.path_importer_cache):
      if path.startswith(pex_file):
        sys.path_importer_cache.pop(path)

    sys.path_hooks.insert(0, pypy_zipimporter_workaround)

  @classmethod
  def _force_local(cls, pex_file, pex_info):
    if pex_info.code_hash is None:
      # Do not support force_local if code_hash is not set. (It should always be set.)
      return pex_file
    explode_dir = os.path.join(pex_info.zip_unsafe_cache, pex_info.code_hash)
    TRACER.log('PEX is not zip safe, exploding to %s' % explode_dir)
    with atomic_directory(explode_dir) as explode_tmp:
      if explode_tmp:
        with TRACER.timed('Unzipping %s' % pex_file):
          with open_zip(pex_file) as pex_zip:
            pex_files = (x for x in pex_zip.namelist()
                         if not x.startswith(pex_builder.BOOTSTRAP_DIR) and
                            not x.startswith(pex_info.internal_cache))
            pex_zip.extractall(explode_tmp, pex_files)
    return explode_dir

  @classmethod
  def _update_module_paths(cls, pex_file):
    bootstrap = Bootstrap.locate()

    # Un-import any modules already loaded from within the .pex file.
    to_reimport = []
    for name, module in reversed(sorted(sys.modules.items())):
      if bootstrap.imported_from_bootstrap(module):
        TRACER.log('Not re-importing module %s from bootstrap.' % module, V=3)
        continue

      pkg_path = getattr(module, '__path__', None)
      if pkg_path and any(os.path.realpath(path_item).startswith(pex_file)
                          for path_item in pkg_path):
        sys.modules.pop(name)
        to_reimport.append((name, pkg_path, True))
      elif name != '__main__':  # The __main__ module is special in python and is not re-importable.
        mod_file = getattr(module, '__file__', None)
        if mod_file and os.path.realpath(mod_file).startswith(pex_file):
          sys.modules.pop(name)
          to_reimport.append((name, mod_file, False))

    # And re-import them from the exploded pex.
    for name, existing_path, is_pkg in to_reimport:
      TRACER.log('Re-importing %s %s loaded via %r from exploded pex.'
                 % ('package' if is_pkg else 'module', name, existing_path))
      reimported_module = importlib.import_module(name)
      if is_pkg:
        for path_item in existing_path:
          # NB: It is not guaranteed that __path__ is a list, it may be a PEP-420 namespace package
          # object which supports a limited mutation API; so we append each item individually.
          reimported_module.__path__.append(path_item)

  @classmethod
  def _write_zipped_internal_cache(cls, zf, pex_info):
    cached_distributions = []
    for distribution_name, dist_digest in pex_info.distributions.items():
      internal_dist_path = '/'.join([pex_info.internal_cache, distribution_name])
      cached_location = os.path.join(pex_info.install_cache, dist_digest, distribution_name)
      dist = CacheHelper.cache_distribution(zf, internal_dist_path, cached_location)
      cached_distributions.append(dist)
    return cached_distributions

  @classmethod
  def _load_internal_cache(cls, pex, pex_info):
    """Possibly cache out the internal cache."""
    internal_cache = os.path.join(pex, pex_info.internal_cache)
    with TRACER.timed('Searching dependency cache: %s' % internal_cache, V=2):
      if len(pex_info.distributions) == 0:
        # We have no .deps to load.
        return

      if os.path.isdir(pex):
        search_path = [os.path.join(internal_cache, dist_chroot)
                       for dist_chroot in os.listdir(internal_cache)]
        internal_env = Environment(search_path=search_path)
        for dist_name in internal_env:
          for dist in internal_env[dist_name]:
            yield dist
      else:
        with open_zip(pex) as zf:
          for dist in cls._write_zipped_internal_cache(zf, pex_info):
            yield dist

  def __init__(self, pex, pex_info, interpreter=None, **kw):
    self._internal_cache = os.path.join(pex, pex_info.internal_cache)
    self._pex = pex
    self._pex_info = pex_info
    self._activated = False
    self._working_set = None
    self._interpreter = interpreter or PythonInterpreter.get()
    self._inherit_path = pex_info.inherit_path
    self._supported_tags = frozenset(self._interpreter.identity.supported_tags)
    self._target_interpreter_env = self._interpreter.identity.env_markers

    # For the bug this works around, see: https://bitbucket.org/pypy/pypy/issues/1686
    # NB: This must be installed early before the underlying pex is loaded in any way.
    if self._interpreter.identity.python_tag.startswith('pp') and zipfile.is_zipfile(self._pex):
      self._install_pypy_zipimporter_workaround(self._pex)

    super(PEXEnvironment, self).__init__(
      search_path=[] if pex_info.inherit_path == 'false' else sys.path,
      platform=self._interpreter.identity.platform_tag,
      **kw
    )
    TRACER.log(
      'E: tags for %r x %r -> %s' % (self.platform, self._interpreter, self._supported_tags),
      V=9
    )

  def _update_candidate_distributions(self, distribution_iter):
    for dist in distribution_iter:
      if self.can_add(dist):
        with TRACER.timed('Adding %s' % dist, V=2):
          self.add(dist)

  def can_add(self, dist):
    filename, ext = os.path.splitext(os.path.basename(dist.location))
    if ext.lower() != '.whl':
      # This supports resolving pex's own vendored distributions which are vendored in directory
      # directory with the project name (`pip/` for pip) and not the corresponding wheel name
      # (`pip-19.3.1-py2.py3-none-any.whl/` for pip). Pex only vendors universal wheels for all
      # platforms it supports at buildtime and runtime so this is always safe.
      return True

    try:
      wheel_name_, wheel_raw_version_, wheel_tags = filename.split('-', 2)
    except ValueError:
      return False

    if self._supported_tags.isdisjoint(tags.parse_tag(wheel_tags)):
      return False

    python_requires = dist_metadata.requires_python(dist)
    if not python_requires:
      return True

    return self._interpreter.identity.version_str in python_requires

  def activate(self):
    if not self._activated:
      with TRACER.timed('Activating PEX virtual environment from %s' % self._pex):
        self._working_set = self._activate()
      self._activated = True

    return self._working_set

  def _resolve(self, working_set, reqs):
    environment = self._target_interpreter_env.copy()
    environment['extra'] = list(set(itertools.chain(*(req.extras for req in reqs))))

    reqs_by_key = OrderedDict()
    for req in reqs:
      if req.marker and not req.marker.evaluate(environment=environment):
        TRACER.log('Skipping activation of `%s` due to environment marker de-selection' % req)
        continue
      reqs_by_key.setdefault(req.key, []).append(req)

    unresolved_reqs = OrderedDict()
    resolveds = OrderedSet()

    # Resolve them one at a time so that we can figure out which ones we need to elide should
    # there be an interpreter incompatibility.
    for key, reqs in reqs_by_key.items():
      with TRACER.timed('Resolving {} from {}'.format(key, reqs), V=2):
        # N.B.: We resolve the bare requirement with no version specifiers since the resolve process
        # used to build this pex already did so. There may be multiple distributions satisfying any
        # particular key (e.g.: a Python 2 specific version and a Python 3 specific version for a
        # multi-python PEX) and we want the working set to pick the most appropriate one.
        req = Requirement.parse(key)
        try:
          resolveds.update(working_set.resolve([req], env=self))
        except DistributionNotFound as e:
          TRACER.log('Failed to resolve a requirement: %s' % e)
          requirers = unresolved_reqs.setdefault(e.req, OrderedSet())
          if e.requirers:
            requirers.update(reqs_by_key[requirer] for requirer in e.requirers)

    if unresolved_reqs:
      TRACER.log('Unresolved requirements:')
      for req in unresolved_reqs:
        TRACER.log('  - %s' % req)

      TRACER.log('Distributions contained within this pex:')
      distributions_by_key = defaultdict(list)
      if not self._pex_info.distributions:
        TRACER.log('  None')
      else:
        for dist_name, dist_digest in self._pex_info.distributions.items():
          TRACER.log('  - %s' % dist_name)
          distribution = DistributionHelper.distribution_from_path(
            path=os.path.join(self._pex_info.install_cache, dist_digest, dist_name)
          )
          distributions_by_key[distribution.as_requirement().key].append(distribution)

      if not self._pex_info.ignore_errors:
        items = []
        for index, (requirement, requirers) in enumerate(unresolved_reqs.items()):
          rendered_requirers = ''
          if requirers:
            rendered_requirers = (
              '\n    Required by:'
              '\n      {requirers}'
            ).format(requirers='\n      '.join(map(str, requirers)))

          items.append(
            '{index: 2d}: {requirement}'
            '{rendered_requirers}'
            '\n    But this pex only contains:'
            '\n      {distributions}'.format(
              index=index + 1,
              requirement=requirement,
              rendered_requirers=rendered_requirers,
              distributions='\n      '.join(os.path.basename(d.location)
                                            for d in distributions_by_key[requirement.key])
            )
          )

        die(
          'Failed to execute PEX file. Needed {platform} compatible dependencies for:\n{items}'
          .format(
            platform=Platform.of_interpreter(self._interpreter),
            items='\n'.join(items)
          )
        )

    return resolveds

  _NAMESPACE_PACKAGE_METADATA_RESOURCE = 'namespace_packages.txt'

  @classmethod
  def _get_namespace_packages(cls, dist):
    if dist.has_metadata(cls._NAMESPACE_PACKAGE_METADATA_RESOURCE):
      return list(dist.get_metadata_lines(cls._NAMESPACE_PACKAGE_METADATA_RESOURCE))
    else:
      return []

  @classmethod
  def declare_namespace_packages(cls, resolved_dists):
    namespace_packages_by_dist = OrderedDict()
    for dist in resolved_dists:
      namespace_packages = cls._get_namespace_packages(dist)
      # NB: Dists can explicitly declare empty namespace packages lists to indicate they have none.
      # We only care about dists with one or more namespace packages though; thus, the guard.
      if namespace_packages:
        namespace_packages_by_dist[dist] = namespace_packages

    if not namespace_packages_by_dist:
      return  # Nothing to do here.

    # When declaring namespace packages, we need to do so with the `setuptools` distribution that
    # will be active in the pex environment at runtime and, as such, care must be taken.
    #
    # Properly behaved distributions will declare a dependency on `setuptools`, in which case we
    # use that (non-vendored) distribution. A side-effect of importing `pkg_resources` from that
    # distribution is that a global `pkg_resources.working_set` will be populated. For various
    # `pkg_resources` distribution discovery functions to work, that global
    # `pkg_resources.working_set` must be built with the `sys.path` fully settled. Since all dists
    # in the dependency set (`resolved_dists`) have already been resolved and added to the
    # `sys.path` we're safe to proceed here.
    #
    # Other distributions (notably `twitter.common.*`) in the wild declare `setuptools`-specific
    # `namespace_packages` but do not properly declare a dependency on `setuptools` which they must
    # use to:
    # 1. Declare `namespace_packages` metadata which we just verified they have with the check
    #    above.
    # 2. Declare namespace packages at runtime via the canonical:
    #    `__import__('pkg_resources').declare_namespace(__name__)`
    #
    # For such distributions we fall back to our vendored version of `setuptools`. This is safe,
    # since we'll only introduce our shaded version when no other standard version is present and
    # even then tear it all down when we hand off from the bootstrap to user code.
    pkg_resources, vendored = _import_pkg_resources()
    if vendored:
      dists = '\n'.join(
        '\n{index}. {dist} namespace packages:\n  {ns_packages}'.format(
          index=index + 1,
          dist=dist.as_requirement(),
          ns_packages='\n  '.join(ns_packages)
        ) for index, (dist, ns_packages) in enumerate(namespace_packages_by_dist.items())
      )
      pex_warnings.warn('The `pkg_resources` package was loaded from a pex vendored version when '
                        'declaring namespace packages defined by:\n{dists}\n\nThese distributions '
                        'should fix their `install_requires` to include `setuptools`'
                        .format(dists=dists))

    for pkg in itertools.chain(*namespace_packages_by_dist.values()):
      if pkg in sys.modules:
        pkg_resources.declare_namespace(pkg)

  def _activate(self):
    pex_file = os.path.realpath(self._pex)

    self._update_candidate_distributions(self._load_internal_cache(pex_file, self._pex_info))

    is_zipped_pex = os.path.isfile(pex_file)
    if not self._pex_info.zip_safe and is_zipped_pex:
      explode_dir = self._force_local(pex_file=pex_file, pex_info=self._pex_info)
      # Force subsequent imports to come from the exploded .pex directory rather than the .pex file.
      TRACER.log('Adding exploded non zip-safe pex to the head of sys.path: %s' % explode_dir)
      sys.path[:] = [path for path in sys.path if pex_file != os.path.realpath(path)]
      sys.path.insert(0, explode_dir)
      self._update_module_paths(pex_file=pex_file)
    elif not any(pex_file == os.path.realpath(path) for path in sys.path):
      TRACER.log('Adding pex %s to the head of sys.path: %s'
                 % ('file' if is_zipped_pex else 'dir', pex_file))
      sys.path.insert(0, pex_file)

    all_reqs = [Requirement.parse(req) for req in self._pex_info.requirements]

    working_set = WorkingSet([])
    resolved = self._resolve(working_set, all_reqs)

    for dist in resolved:
      with TRACER.timed('Activating %s' % dist, V=2):
        working_set.add(dist)

        if self._inherit_path == "fallback":
          # Prepend location to sys.path.
          #
          # This ensures that bundled versions of libraries will be used before system-installed
          # versions, in case something is installed in both, helping to favor hermeticity in
          # the case of non-hermetic PEX files (i.e. those with inherit_path=True).
          #
          # If the path is not already in sys.path, site.addsitedir will append (not prepend)
          # the path to sys.path. But if the path is already in sys.path, site.addsitedir will
          # leave sys.path unmodified, but will do everything else it would do. This is not part
          # of its advertised contract (which is very vague), but has been verified to be the
          # case by inspecting its source for both cpython 2.7 and cpython 3.7.
          sys.path.insert(0, dist.location)
        else:
          sys.path.append(dist.location)

        with TRACER.timed('Adding sitedir', V=2):
          site.addsitedir(dist.location)

    return working_set



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import shlex
import subprocess
import sys
from distutils import log
from distutils.core import Command

from pex.bin.pex import configure_clp
from pex.common import die
from pex.compatibility import ConfigParser, StringIO, string, to_unicode
from pex.interpreter import PythonInterpreter


# Suppress checkstyle violations due to distutils command requirements.
class bdist_pex(Command):  # noqa
  @staticmethod
  def get_log_level():
    # A hack to get the existing distutils logging level.
    existing_level = log.set_threshold(log.INFO)
    log.set_threshold(existing_level)

    return existing_level

  description = "create a PEX file from a source distribution"  # noqa

  user_options = [  # noqa
      ('bdist-all', None, 'pexify all defined entry points'),
      ('bdist-dir=', None, 'the directory into which pexes will be written, default: dist.'),
      ('pex-args=', None, 'additional arguments to the pex tool'),
  ]

  boolean_options = [  # noqa
    'bdist-all',
  ]

  def initialize_options(self):
    self.bdist_all = False
    self.bdist_dir = None
    self.pex_args = ''

  def finalize_options(self):
    self.pex_args = shlex.split(self.pex_args)

  def parse_entry_points(self):
    def parse_entry_point_name(entry_point):
      script_name = entry_point.split('=', 1)[0]
      return script_name.strip()

    raw_entry_points = self.distribution.entry_points

    if isinstance(raw_entry_points, string):
      parser = ConfigParser()
      parser.readfp(StringIO(to_unicode(raw_entry_points)))
      if parser.has_section('console_scripts'):
        return tuple(parser.options('console_scripts'))
    elif isinstance(raw_entry_points, dict):
      try:
        return tuple(parse_entry_point_name(script)
                     for script in raw_entry_points.get('console_scripts', []))
      except ValueError:
        pass
    elif raw_entry_points is not None:
      die('When entry_points is provided, it must be a string or dict.')

    return ()

  def run(self):
    parser = configure_clp()
    options, reqs = parser.parse_args(self.pex_args)

    if options.entry_point or options.script or options.pex_name:
      die('Must not specify entry point, script or output file to --pex-args, given: {}'
          .format(' '.join(self.pex_args)))

    name = self.distribution.get_name()
    version = self.distribution.get_version()

    package_dir = os.path.dirname(os.path.realpath(os.path.expanduser(
      self.distribution.script_name)))
    if self.bdist_dir is None:
      self.bdist_dir = os.path.join(package_dir, 'dist')

    console_scripts = self.parse_entry_points()

    pex_specs = []
    if self.bdist_all:
      # Write all entry points into unversioned pex files.
      pex_specs.extend((script_name, os.path.join(self.bdist_dir, script_name))
                       for script_name in console_scripts)
    else:
      target = os.path.join(self.bdist_dir, name + '-' + version + '.pex')
      pex_specs.append((name if name in console_scripts else None, target))

    args = ['-m', 'pex', package_dir] + reqs + self.pex_args
    if self.get_log_level() < log.INFO and options.verbosity == 0:
      args.append('-v')

    for script_name, target in pex_specs:
      pex_cmd = args + ['--output-file', target]
      if script_name:
        log.info('Writing %s to %s' % (script_name, target))
        pex_cmd += ['--script', script_name]
      else:
        # The package has no namesake entry point, so build an environment pex.
        log.info('Writing environment pex into %s' % target)

      cmd, process = PythonInterpreter.get().open_process(
        args=pex_cmd,
        stderr=subprocess.PIPE,

        # In order for code to run to here, pex is on the sys.path - make sure to propagate the
        # sys.path so the subprocess can find us.
        pythonpath=sys.path
      )
      _, stderr = process.communicate()
      result = process.returncode
      if result != 0:
        die('Failed to create pex via {}:\n{}'.format(' '.join(cmd), stderr.decode('utf-8')),
            result)



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import sys
from textwrap import dedent

import pytest

from pex.interpreter import PythonInterpreter
from pex.interpreter_constraints import UnsatisfiableInterpreterConstraintsError
from pex.pex_bootstrapper import iter_compatible_interpreters
from pex.testing import PY27, PY35, PY36, ensure_python_interpreter


def find_interpreters(path, *constraints):
  return [interp.binary for interp in
          iter_compatible_interpreters(path=os.pathsep.join(path),
                                       compatibility_constraints=constraints)]


def test_find_compatible_interpreters():
  py27 = ensure_python_interpreter(PY27)
  py35 = ensure_python_interpreter(PY35)
  py36 = ensure_python_interpreter(PY36)
  path = [py27, py35, py36]

  assert [py35, py36] == find_interpreters(path, '>3')
  assert [py27] == find_interpreters(path, '<3')

  assert [py36] == find_interpreters(path, '>{}'.format(PY35))
  assert [py35] == find_interpreters(path, '>{}, <{}'.format(PY27, PY36))
  assert [py36] == find_interpreters(path, '>=3.6')

  with pytest.raises(UnsatisfiableInterpreterConstraintsError):
    find_interpreters(path, '<2')

  with pytest.raises(UnsatisfiableInterpreterConstraintsError):
    find_interpreters(path, '>4')

  with pytest.raises(UnsatisfiableInterpreterConstraintsError):
    find_interpreters(path, '>{}, <{}'.format(PY27, PY35))

  # All interpreters on PATH including whatever interpreter is currently running.
  all_known_interpreters = set(PythonInterpreter.all())
  all_known_interpreters.add(PythonInterpreter.get())

  interpreters = set(iter_compatible_interpreters(compatibility_constraints=['<3']))
  i_rendered = '\n      '.join(sorted(map(repr, interpreters)))
  aki_rendered = '\n      '.join(sorted(map(repr, all_known_interpreters)))
  assert interpreters.issubset(all_known_interpreters), dedent(
    """
    interpreters '<3':
      {interpreters}

    all known interpreters:
      {all_known_interpreters}
    """.format(interpreters=i_rendered, all_known_interpreters=aki_rendered)
  )


def test_find_compatible_interpreters_bias_current():
  py36 = ensure_python_interpreter(PY36)
  assert [os.path.realpath(sys.executable), py36] == find_interpreters([py36, sys.executable])
  assert [os.path.realpath(sys.executable), py36] == find_interpreters([sys.executable, py36])



# Copyright 2019 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.interpreter import PythonInterpreter
from pex.platforms import Platform


class DistributionTarget(object):
  """Represents the target of a python distribution."""

  @classmethod
  def current(cls):
    return cls(interpreter=None, platform=None)

  @classmethod
  def for_interpreter(cls, interpreter):
    return cls(interpreter=interpreter, platform=None)

  @classmethod
  def for_platform(cls, platform):
    return cls(interpreter=None, platform=platform)

  def __init__(self, interpreter=None, platform=None):
    self._interpreter = interpreter
    self._platform = platform

  @property
  def is_foreign(self):
    if self._platform is None:
      return False
    return self._platform != Platform.of_interpreter(self._interpreter)

  def get_interpreter(self):
    return self._interpreter or PythonInterpreter.get()

  def get_platform(self):
    return self._platform or Platform.current()

  def requirement_applies(self, requirement):
    """Determines if the given requirement applies to this distribution target.

    :param requirement: The requirement to evaluate.
    :type requirement: :class:`pex.third_party.pkg_resources.Requirement`
    :rtype: bool
    """
    if not requirement.marker:
      return True

    if self._interpreter is None:
      return True

    return requirement.marker.evaluate(self._interpreter.identity.env_markers)

  @property
  def id(self):
    """A unique id for a resolve target suitable as a path name component.

    :rtype: str
    """
    if self._platform is None:
      interpreter = self.get_interpreter()
      return '{python}-{abi}'.format(python=interpreter.identity.python_tag,
                                     abi=interpreter.identity.abi_tag)
    else:
      return str(self._platform)

  def __repr__(self):
    if self._platform is None:
      return '{}(interpreter={!r})'.format(self.__class__.__name__, self.get_interpreter())
    else:
      return '{}(platform={!r})'.format(self.__class__.__name__, self._platform)

  def _tup(self):
    return self._interpreter, self._platform

  def __eq__(self, other):
    if type(other) is not type(self):
      return NotImplemented
    return self._tup() == other._tup()

  def __hash__(self):
    return hash(self._tup())



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

"""
The pex.bin.pex utility builds PEX environments and .pex files specified by
sources, requirements and their dependencies.
"""

from __future__ import absolute_import, print_function

import os
import sys
from optparse import OptionGroup, OptionParser, OptionValueError
from textwrap import TextWrapper

from pex import pex_warnings
from pex.common import die, safe_delete, safe_mkdtemp
from pex.interpreter import PythonInterpreter
from pex.interpreter_constraints import (
    UnsatisfiableInterpreterConstraintsError,
    validate_constraints
)
from pex.jobs import DEFAULT_MAX_JOBS
from pex.pex import PEX
from pex.pex_bootstrapper import iter_compatible_interpreters
from pex.pex_builder import PEXBuilder
from pex.platforms import Platform
from pex.resolver import Unsatisfiable, parsed_platform, resolve_multi
from pex.tracer import TRACER
from pex.variables import ENV, Variables
from pex.version import __version__

CANNOT_SETUP_INTERPRETER = 102
INVALID_OPTIONS = 103


def log(msg, V=0):
  if V != 0:
    print(msg, file=sys.stderr)


def parse_bool(option, opt_str, _, parser):
  setattr(parser.values, option.dest, not opt_str.startswith('--no'))


def increment_verbosity(option, opt_str, _, parser):
  verbosity = getattr(parser.values, option.dest, 0)
  setattr(parser.values, option.dest, verbosity + 1)


def process_disable_cache(option, option_str, option_value, parser):
  setattr(parser.values, option.dest, None)


class PyPiSentinel(object):
  def __str__(self):
    return 'https://pypi.org/simple'


_PYPI = PyPiSentinel()


def process_pypi_option(option, option_str, option_value, parser):
  if option_str.startswith('--no'):
    setattr(parser.values, option.dest, [])
  else:
    indexes = getattr(parser.values, option.dest, [])
    if _PYPI not in indexes:
      indexes.append(_PYPI)
    setattr(parser.values, option.dest, indexes)


def process_find_links(option, option_str, option_value, parser):
  find_links = getattr(parser.values, option.dest, [])
  if option_value not in find_links:
    find_links.append(option_value)
  setattr(parser.values, option.dest, find_links)


def process_index_url(option, option_str, option_value, parser):
  indexes = getattr(parser.values, option.dest, [])
  if option_value not in indexes:
    indexes.append(option_value)
  setattr(parser.values, option.dest, indexes)


_DEFAULT_MANYLINUX_STANDARD = 'manylinux2014'


def process_manylinux(option, option_str, option_value, parser):
  if option_str.startswith('--no'):
    setattr(parser.values, option.dest, None)
  elif option_value.startswith('manylinux'):
    setattr(parser.values, option.dest, option_value)
  else:
    raise OptionValueError('Please specify a manylinux standard; ie: --manylinux=manylinux1. '
                           'Given {}'.format(option_value))


def process_transitive(option, option_str, option_value, parser):
  transitive = option_str == '--transitive'
  setattr(parser.values, option.dest, transitive)


def print_variable_help(option, option_str, option_value, parser):
  for variable_name, variable_type, variable_help in Variables.iter_help():
    print('\n%s: %s\n' % (variable_name, variable_type))
    for line in TextWrapper(initial_indent=' ' * 4, subsequent_indent=' ' * 4).wrap(variable_help):
      print(line)
  sys.exit(0)


def process_platform(option, option_str, option_value, parser):
  platforms = getattr(parser.values, option.dest, [])
  try:
    platforms.append(parsed_platform(option_value))
  except Platform.InvalidPlatformError as e:
    raise OptionValueError("The {} option is invalid:\n{}"
                           .format(option_str, e))


def configure_clp_pex_resolution(parser):
  group = OptionGroup(
      parser,
      'Resolver options',
      'Tailor how to find, resolve and translate the packages that get put into the PEX '
      'environment.')

  group.add_option(
      '--pypi', '--no-pypi', '--no-index',
      action='callback',
      dest='indexes',
      default=[_PYPI],
      callback=process_pypi_option,
      help='Whether to use pypi to resolve dependencies; Default: use pypi')

  group.add_option(
    '--pex-path',
    dest='pex_path',
    type=str,
    default=None,
    help='A colon separated list of other pex files to merge into the runtime environment.')

  group.add_option(
      '-f', '--find-links', '--repo',
      metavar='PATH/URL',
      action='callback',
      default=[],
      dest='find_links',
      callback=process_find_links,
      type=str,
      help='Additional repository path (directory or URL) to look for requirements.')

  group.add_option(
      '-i', '--index', '--index-url',
      metavar='URL',
      action='callback',
      dest='indexes',
      callback=process_index_url,
      type=str,
      help='Additional cheeseshop indices to use to satisfy requirements.')

  group.add_option(
    '--pre', '--no-pre',
    dest='allow_prereleases',
    default=False,
    action='callback',
    callback=parse_bool,
    help='Whether to include pre-release and development versions of requirements; '
         'Default: only stable versions are used, unless explicitly requested')

  group.add_option(
      '--disable-cache',
      dest='disable_cache',
      default=False,
      action='store_true',
      help='Disable caching in the pex tool entirely.')

  group.add_option(
      '--cache-dir',
      dest='cache_dir',
      default=None,
      help='DEPRECATED: Use --pex-root instead. '
           'The local cache directory to use for speeding up requirement '
           'lookups. [Default: {}]'.format(ENV.PEX_ROOT))

  group.add_option(
      '--wheel', '--no-wheel', '--no-use-wheel',
      dest='use_wheel',
      default=True,
      action='callback',
      callback=parse_bool,
      help='Whether to allow wheel distributions; Default: allow wheels')

  group.add_option(
      '--build', '--no-build',
      dest='build',
      default=True,
      action='callback',
      callback=parse_bool,
      help='Whether to allow building of distributions from source; Default: allow builds')

  group.add_option(
      '--manylinux', '--no-manylinux', '--no-use-manylinux',
      dest='manylinux',
      type=str,
      default=_DEFAULT_MANYLINUX_STANDARD,
      action='callback',
      callback=process_manylinux,
      help=('Whether to allow resolution of manylinux wheels for linux target '
            'platforms; Default: allow manylinux wheels compatible with {}'
            .format(_DEFAULT_MANYLINUX_STANDARD)))

  group.add_option(
    '--transitive', '--no-transitive', '--intransitive',
    dest='transitive',
    default=True,
    action='callback',
    callback=process_transitive,
    help='Whether to transitively resolve requirements. Default: True')

  group.add_option(
    '-j', '--jobs',
    metavar='JOBS',
    dest='max_parallel_jobs',
    type=int,
    default=DEFAULT_MAX_JOBS,
    help='The maximum number of parallel jobs to use when resolving, building and installing '
         'distributions. You might want to increase the maximum number of parallel jobs to '
         'potentially improve the latency of the pex creation process at the expense of other'
         'processes on your system. [Default: %default]')

  parser.add_option_group(group)


def configure_clp_pex_options(parser):
  group = OptionGroup(
      parser,
      'PEX output options',
      'Tailor the behavior of the emitted .pex file if -o is specified.')

  group.add_option(
      '--zip-safe', '--not-zip-safe',
      dest='zip_safe',
      default=True,
      action='callback',
      callback=parse_bool,
      help='Whether or not the sources in the pex file are zip safe.  If they are not zip safe, '
           'they will be written to disk prior to execution. Also see --unzip which will cause the '
           'complete pex file, including dependencies, to be unzipped.'
           '[Default: zip safe.]')

  group.add_option(
    '--unzip', '--no-unzip',
    dest='unzip',
    default=False,
    action='callback',
    callback=parse_bool,
    help='Whether or not the pex file should be unzipped before executing it. If the pex file will '
         'be run multiple times under a stable runtime PEX_ROOT the unzipping will only be '
         'performed once and subsequent runs will enjoy lower startup latency. '
         '[Default: do not unzip.]')

  group.add_option(
      '--always-write-cache',
      dest='always_write_cache',
      default=False,
      action='store_true',
      help='Always write the internally cached distributions to disk prior to invoking '
           'the pex source code.  This can use less memory in RAM constrained '
           'environments. [Default: %default]')

  group.add_option(
      '--ignore-errors',
      dest='ignore_errors',
      default=False,
      action='store_true',
      help='Ignore requirement resolution solver errors when building pexes and later invoking '
           'them. [Default: %default]')

  group.add_option(
      '--inherit-path',
      dest='inherit_path',
      default='false',
      action='store',
      choices=['false', 'fallback', 'prefer'],
      help='Inherit the contents of sys.path (including site-packages, user site-packages and '
           'PYTHONPATH) running the pex. Possible values: false (does not inherit sys.path), '
           'fallback (inherits sys.path after packaged dependencies), prefer (inherits sys.path '
           'before packaged dependencies), No value (alias for prefer, for backwards '
           'compatibility). [Default: %default]')

  group.add_option(
      '--compile', '--no-compile',
      dest='compile',
      default=False,
      action='callback',
      callback=parse_bool,
      help='Compiling means that the built pex will include .pyc files, which will result in '
           'slightly faster startup performance. However, compiling means that the generated pex '
           'likely will not be reproducible, meaning that if you were to run `./pex -o` with the '
           'same inputs then the new pex would not be byte-for-byte identical to the original.')

  group.add_option(
      '--use-system-time', '--no-use-system-time',
      dest='use_system_time',
      default=False,
      action='callback',
      callback=parse_bool,
      help='Use the current system time to generate timestamps for the new pex. Otherwise, Pex '
           'will use midnight on January 1, 1980. By using system time, the generated pex '
           'will not be reproducible, meaning that if you were to run `./pex -o` with the '
           'same inputs then the new pex would not be byte-for-byte identical to the original.')

  group.add_option(
      '--runtime-pex-root',
      dest='runtime_pex_root',
      default=None,
      help='Specify the pex root to be used in the generated .pex file. [Default: ~/.pex]')

  group.add_option(
      '--strip-pex-env', '--no-strip-pex-env',
      dest='strip_pex_env',
      default=True,
      action='callback',
      callback=parse_bool,
      help='Strip all `PEX_*` environment variables used to control the pex runtime before handing '
           'off control to the pex entrypoint. You might want to set this to `False` if the new '
           'pex executes other pexes (or the Pex CLI itself and you want the executed pex to be '
           'controllable via `PEX_*` environment variables. [Default: %default]')

  parser.add_option_group(group)


def configure_clp_pex_environment(parser):
  group = OptionGroup(
      parser,
      'PEX environment options',
      'Tailor the interpreter and platform targets for the PEX environment.')

  group.add_option(
      '--python',
      dest='python',
      default=[],
      type='str',
      action='append',
      help='The Python interpreter to use to build the pex.  Either specify an explicit '
           'path to an interpreter, or specify a binary accessible on $PATH. This option '
           'can be passed multiple times to create a multi-interpreter compatible pex. '
           'Default: Use current interpreter.')

  group.add_option(
    '--interpreter-constraint',
    dest='interpreter_constraint',
    default=[],
    type='str',
    action='append',
    help='Constrain the selected Python interpreter. Specify with Requirement-style syntax, '
         'e.g. "CPython>=2.7,<3" (A CPython interpreter with version >=2.7 AND version <3) '
         'or "PyPy" (A pypy interpreter of any version). This argument may be repeated multiple '
         'times to OR the constraints.')

  group.add_option(
    '--rcfile',
    dest='rc_file',
    default=None,
    help='An additional path to a pexrc file to read during configuration parsing. '
         'Used primarily for testing.')

  group.add_option(
      '--python-shebang',
      dest='python_shebang',
      default=None,
      help='The exact shebang (#!...) line to add at the top of the PEX file minus the '
           '#!.  This overrides the default behavior, which picks an environment python '
           'interpreter compatible with the one used to build the PEX file.')

  group.add_option(
      '--platform',
      dest='platforms',
      default=[],
      type=str,
      action='callback',
      callback=process_platform,
      help='The platform for which to build the PEX. This option can be passed multiple times '
           'to create a multi-platform pex. To use wheels for specific interpreter/platform tags'
           ', you can append them to the platform with hyphens like: PLATFORM-IMPL-PYVER-ABI '
           '(e.g. "linux_x86_64-cp-27-cp27mu", "macosx_10.12_x86_64-cp-36-cp36m") PLATFORM is '
           'the host platform e.g. "linux-x86_64", "macosx-10.12-x86_64", etc". IMPL is the '
           'python implementation abbreviation (e.g. "cp", "pp", "jp"). PYVER is a two-digit '
           'string representing the python version (e.g. "27", "36"). ABI is the ABI tag '
           '(e.g. "cp36m", "cp27mu", "abi3", "none"). Default: current platform.')

  parser.add_option_group(group)


def configure_clp_pex_entry_points(parser):
  group = OptionGroup(
      parser,
      'PEX entry point options',
      'Specify what target/module the PEX should invoke if any.')

  group.add_option(
      '-m', '-e', '--entry-point',
      dest='entry_point',
      metavar='MODULE[:SYMBOL]',
      default=None,
      help='Set the entry point to module or module:symbol.  If just specifying module, pex '
           'behaves like python -m, e.g. python -m SimpleHTTPServer.  If specifying '
           'module:symbol, pex imports that symbol and invokes it as if it were main.')

  group.add_option(
      '-c', '--script', '--console-script',
      dest='script',
      default=None,
      metavar='SCRIPT_NAME',
      help='Set the entry point as to the script or console_script as defined by a any of the '
           'distributions in the pex.  For example: "pex -c fab fabric" or "pex -c mturk boto".')

  group.add_option(
      '--validate-entry-point',
      dest='validate_ep',
      default=False,
      action='store_true',
      help='Validate the entry point by importing it in separate process. Warning: this could have '
           'side effects. For example, entry point `a.b.c:m` will translate to '
           '`from a.b.c import m` during validation. [Default: %default]')

  parser.add_option_group(group)


def configure_clp():
  usage = (
      '%prog [-o OUTPUT.PEX] [options] [-- arg1 arg2 ...]\n\n'
      '%prog builds a PEX (Python Executable) file based on the given specifications: '
      'sources, requirements, their dependencies and other options.')

  parser = OptionParser(usage=usage, version='%prog {}'.format(__version__))
  configure_clp_pex_resolution(parser)
  configure_clp_pex_options(parser)
  configure_clp_pex_environment(parser)
  configure_clp_pex_entry_points(parser)

  parser.add_option(
      '-o', '--output-file',
      dest='pex_name',
      default=None,
      help='The name of the generated .pex file: Omiting this will run PEX '
           'immediately and not save it to a file.')

  parser.add_option(
      '-p', '--preamble-file',
      dest='preamble_file',
      metavar='FILE',
      default=None,
      type=str,
      help='The name of a file to be included as the preamble for the generated .pex file')

  parser.add_option(
      '-D', '--sources-directory',
      dest='sources_directory',
      metavar='DIR',
      default=[],
      type=str,
      action='append',
      help='Add sources directory to be packaged into the generated .pex file.'
           '  This option can be used multiple times.')

  parser.add_option(
      '-R', '--resources-directory',
      dest='resources_directory',
      metavar='DIR',
      default=[],
      type=str,
      action='append',
      help='Add resources directory to be packaged into the generated .pex file.'
           '  This option can be used multiple times.')

  parser.add_option(
      '-r', '--requirement',
      dest='requirement_files',
      metavar='FILE',
      default=[],
      type=str,
      action='append',
      help='Add requirements from the given requirements file.  This option can be used multiple '
           'times.')

  parser.add_option(
      '--constraints',
      dest='constraint_files',
      metavar='FILE',
      default=[],
      type=str,
      action='append',
      help='Add constraints from the given constraints file.  This option can be used multiple '
           'times.')

  parser.add_option(
    '--requirements-pex',
    dest='requirements_pexes',
    metavar='FILE',
    default=[],
    type=str,
    action='append',
    help='Add requirements from the given .pex file.  This option can be used multiple times.')

  parser.add_option(
      '-v',
      dest='verbosity',
      default=0,
      action='callback',
      callback=increment_verbosity,
      help='Turn on logging verbosity, may be specified multiple times.')

  parser.add_option(
      '--emit-warnings', '--no-emit-warnings',
      dest='emit_warnings',
      action='callback',
      callback=parse_bool,
      default=True,
      help='Emit runtime UserWarnings on stderr. If false, only emit them when PEX_VERBOSE is set.'
           'Default: emit user warnings to stderr')

  parser.add_option(
      '--pex-root',
      dest='pex_root',
      default=None,
      help='Specify the pex root used in this invocation of pex. '
           '[Default: {}]'.format(ENV.PEX_ROOT))

  parser.add_option(
      '--help-variables',
      action='callback',
      callback=print_variable_help,
      help='Print out help about the various environment variables used to change the behavior of '
           'a running PEX file.')

  return parser


def _safe_link(src, dst):
  try:
    os.unlink(dst)
  except OSError:
    pass
  os.symlink(src, dst)


def build_pex(reqs, options, cache=None):
  interpreters = None  # Default to the current interpreter.

  # NB: options.python and interpreter constraints cannot be used together.
  if options.python:
    with TRACER.timed('Resolving interpreters', V=2):
      def to_python_interpreter(full_path_or_basename):
        if os.path.isfile(full_path_or_basename):
          return PythonInterpreter.from_binary(full_path_or_basename)
        else:
          interpreter = PythonInterpreter.from_env(full_path_or_basename)
          if interpreter is None:
            die('Failed to find interpreter: %s' % full_path_or_basename)
          return interpreter

      interpreters = [to_python_interpreter(interp) for interp in options.python]
  elif options.interpreter_constraint:
    with TRACER.timed('Resolving interpreters', V=2):
      constraints = options.interpreter_constraint
      validate_constraints(constraints)
      if options.rc_file or not ENV.PEX_IGNORE_RCFILES:
        rc_variables = Variables.from_rc(rc=options.rc_file)
        pex_python_path = rc_variables.get('PEX_PYTHON_PATH', None)
      else:
        pex_python_path = None
      try:
        interpreters = list(iter_compatible_interpreters(pex_python_path, constraints))
      except UnsatisfiableInterpreterConstraintsError as e:
        die(e.create_message('Could not find a compatible interpreter.'), CANNOT_SETUP_INTERPRETER)

  try:
    with open(options.preamble_file) as preamble_fd:
      preamble = preamble_fd.read()
  except TypeError:
    # options.preamble_file is None
    preamble = None

  interpreter = min(interpreters) if interpreters else None

  pex_builder = PEXBuilder(path=safe_mkdtemp(), interpreter=interpreter, preamble=preamble)

  def walk_and_do(fn, src_dir):
    src_dir = os.path.normpath(src_dir)
    for root, dirs, files in os.walk(src_dir):
      for f in files:
        src_file_path = os.path.join(root, f)
        dst_path = os.path.relpath(src_file_path, src_dir)
        fn(src_file_path, dst_path)

  for directory in options.sources_directory:
    walk_and_do(pex_builder.add_source, directory)

  for directory in options.resources_directory:
    walk_and_do(pex_builder.add_resource, directory)

  pex_info = pex_builder.info
  pex_info.zip_safe = options.zip_safe
  pex_info.unzip = options.unzip
  pex_info.pex_path = options.pex_path
  pex_info.always_write_cache = options.always_write_cache
  pex_info.ignore_errors = options.ignore_errors
  pex_info.emit_warnings = options.emit_warnings
  pex_info.inherit_path = options.inherit_path
  pex_info.pex_root = options.runtime_pex_root
  pex_info.strip_pex_env = options.strip_pex_env
  if options.interpreter_constraint:
    for ic in options.interpreter_constraint:
      pex_builder.add_interpreter_constraint(ic)

  # NB: `None` means use the default (pypi) index, `[]` means use no indexes.
  indexes = None
  if options.indexes != [_PYPI] and options.indexes is not None:
    indexes = [str(index) for index in options.indexes]

  for requirements_pex in options.requirements_pexes:
    pex_builder.add_from_requirements_pex(requirements_pex)

  with TRACER.timed('Resolving distributions ({})'.format(reqs + options.requirement_files)):
    try:
      resolveds = resolve_multi(requirements=reqs,
                                requirement_files=options.requirement_files,
                                constraint_files=options.constraint_files,
                                allow_prereleases=options.allow_prereleases,
                                transitive=options.transitive,
                                interpreters=interpreters,
                                platforms=options.platforms,
                                indexes=indexes,
                                find_links=options.find_links,
                                cache=cache,
                                build=options.build,
                                use_wheel=options.use_wheel,
                                compile=options.compile,
                                manylinux=options.manylinux,
                                max_parallel_jobs=options.max_parallel_jobs,
                                ignore_errors=options.ignore_errors)

      for resolved_dist in resolveds:
        log('  %s -> %s' % (resolved_dist.requirement, resolved_dist.distribution),
            V=options.verbosity)
        pex_builder.add_distribution(resolved_dist.distribution)
        pex_builder.add_requirement(resolved_dist.requirement)
    except Unsatisfiable as e:
      die(e)

  if options.entry_point and options.script:
    die('Must specify at most one entry point or script.', INVALID_OPTIONS)

  if options.entry_point:
    pex_builder.set_entry_point(options.entry_point)
  elif options.script:
    pex_builder.set_script(options.script)

  if options.python_shebang:
    pex_builder.set_shebang(options.python_shebang)

  return pex_builder


def transform_legacy_arg(arg):
  # inherit-path used to be a boolean arg (so either was absent, or --inherit-path)
  # Now it takes a string argument, so --inherit-path is invalid.
  # Fix up the args we're about to parse to preserve backwards compatibility.
  if arg == '--inherit-path':
    return '--inherit-path=prefer'
  return arg


def _compatible_with_current_platform(platforms):
  if not platforms:
    return True
  current_platforms = {None, Platform.current()}
  return current_platforms.intersection(platforms)


def main(args=None):
  args = args[:] if args else sys.argv[1:]
  args = [transform_legacy_arg(arg) for arg in args]
  parser = configure_clp()

  try:
    separator = args.index('--')
    args, cmdline = args[:separator], args[separator + 1:]
  except ValueError:
    args, cmdline = args, []

  options, reqs = parser.parse_args(args=args)

  if options.cache_dir:
    pex_warnings.warn('The --cache-dir option is deprecated, use --pex-root instead.')
    if options.pex_root and options.cache_dir != options.pex_root:
      die('Both --cache-dir and --pex-root were passed with conflicting values. '
          'Just set --pex-root.')

  if options.disable_cache:
    def warn_ignore_pex_root(set_via):
      pex_warnings.warn('The pex root has been set via {via} but --disable-cache is also set. '
                        'Ignoring {via} and disabling caches.'.format(via=set_via))

    if options.cache_dir:
      warn_ignore_pex_root('--cache-dir')
    elif options.pex_root:
      warn_ignore_pex_root('--pex-root')
    elif os.environ.get('PEX_ROOT'):
      warn_ignore_pex_root('PEX_ROOT')

    pex_root = safe_mkdtemp()
  else:
    pex_root = options.cache_dir or options.pex_root or ENV.PEX_ROOT

  if options.python and options.interpreter_constraint:
    die('The "--python" and "--interpreter-constraint" options cannot be used together.')

  with ENV.patch(PEX_VERBOSE=str(options.verbosity), PEX_ROOT=pex_root) as patched_env:
    with TRACER.timed('Building pex'):
      pex_builder = build_pex(reqs, options, cache=ENV.PEX_ROOT)

    pex_builder.freeze(bytecode_compile=options.compile)
    pex = PEX(pex_builder.path(),
              interpreter=pex_builder.interpreter,
              verify_entry_point=options.validate_ep)

    if options.pex_name is not None:
      log('Saving PEX file to %s' % options.pex_name, V=options.verbosity)
      tmp_name = options.pex_name + '~'
      safe_delete(tmp_name)
      pex_builder.build(
        tmp_name,
        bytecode_compile=options.compile,
        deterministic_timestamp=not options.use_system_time
      )
      os.rename(tmp_name, options.pex_name)
    else:
      if not _compatible_with_current_platform(options.platforms):
        log('WARNING: attempting to run PEX with incompatible platforms!', V=1)
        log('Running on platform {} but built for {}'
            .format(Platform.current(), ', '.join(map(str, options.platforms))), V=1)

      log('Running PEX file at %s with args %s' % (pex_builder.path(), cmdline),
          V=options.verbosity)
      sys.exit(pex.run(args=list(cmdline), env=patched_env))


if __name__ == '__main__':
  main()



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import re


class UnsupportedLine(Exception):
  pass


def _get_parameter(line):
  sline = line.split('=')
  if len(sline) != 2:
    sline = line.split()
  if len(sline) != 2:
    raise UnsupportedLine('Unrecognized line format: %s' % line)
  return sline[1]


def local_project_from_requirement(requirement, relpath=None):
  """Return the absolute path for a local project requirement.

  :param str requirement: The requirement string to parse.
  :param str relpath: The base path to measure the local project path from.
  :return: The absolute path of the local project requirement or `None` if the requirement string
           does not represent a local project requirement.
  """

  relpath = relpath or os.getcwd()

  # Strip any extras that may be present. e.g. given: `./local/setup-py-project[foo]`
  # produce: `./local/setup-py-project`
  maybe_dir = re.sub(r'\[(?#extras).*\]$', '', requirement)
  maybe_abs_dir = maybe_dir if os.path.isabs(maybe_dir) else os.path.join(relpath, maybe_dir)
  if any(os.path.isfile(os.path.join(maybe_abs_dir, f)) for f in ('setup.py', 'pyproject.toml')):
    return maybe_abs_dir


# Process lines in the requirements.txt format as defined here:
# https://pip.pypa.io/en/latest/reference/pip_install.html#requirements-file-format
# Note that we're only interested in requirement specifiers that look like local directories and not
# any pip options which we'll let pip handle.
def _iter_local_projects(lines, relpath=None):
  relpath = relpath or os.getcwd()

  for line in lines:
    line = line.strip()
    if not line or line.startswith('#'):
      continue
    elif line.startswith(('-r ', '--requirement')):
      path = os.path.join(relpath, _get_parameter(line))
      for local_project in local_projects_from_requirement_file(path):
        yield local_project
    elif not line.startswith('-'):
      local_project = local_project_from_requirement(line, relpath=relpath)
      if local_project:
        yield local_project


def local_projects_from_requirement_file(filename):
  """Return a list of local project absolute paths from a requirements.txt file.

  :param filename: The filename of the requirements file.
  """

  relpath = os.path.dirname(filename)
  with open(filename, 'r') as fp:
    return _iter_local_projects(fp.readlines(), relpath=relpath)
